"""
Class for a neural network Layer.
"""
import sys
import numpy as np
from scipy.stats import norm


class Layer:
    """
    Class for a neural network Layer.

    Method and notation convention:
        Aggarwal, Charu C. "Neural networks and deep learning."
        Springer 10 (2018): 978-3.
    """

    def __init__(self, n_neurons, r, n_layers, activation, loss, bias=False,
                 batch_size=1, lamb=0.0, on_gpu=False,
                 n_softmax=0, **kwargs):
        """
        Create a Layer object.

        Parameters
        ----------
        n_neurons : int, optional
            The number of neurons per hidden layer. The default is 16.
        r : int
            The layer index, 0 being the input layer and n_layers the output layer.
        n_layers : int, optional
            The number of layers, not counting the input layer. The default is 2.
        activation : string, optional
            The name of the activation function of the hidden layers.
            The default is 'tanh'.
        loss : string, optional
            The name of the loss function. The default is 'squared'.
        bias : boolean, optional
            Use a bias neuron. The default is True.
        batch_size : int, optional
            The size of the mini batch. The default is 1.
        lamb : float, optional
            L2 weight regularization parameter. The default is 0.0.
        on_gpu : boolean, optional
            Train the neural network on a GPU using cupy. NOT IMPLEMENTED IN THIS VERSION.
            The default is False.
        n_softmax : int, optional
            The number of softmax layers attached to the output. The default is 0.

        Returns
        -------
        None.

        """

        self.n_neurons = n_neurons
        self.r = r
        self.n_layers = n_layers
        self.activation = activation
        self.loss = loss
        self.bias = bias
        self.batch_size = batch_size
        self.lamb = lamb
        self.n_softmax = n_softmax

        # #use either numpy or cupy via xp based on the on_gpu flag
        # global xp
        # if on_gpu == False:
        #     import numpy as xp
        # else:
        #     import cupy as xp

        self.on_gpu = False

        if self.bias:
            self.n_bias = 1
        else:
            self.n_bias = 0

        # the number of classes in each softmax layer
        if self.n_softmax > 0:
            self.n_bins = int(self.n_neurons / self.n_softmax)

        self.a = np.zeros([n_neurons, batch_size])
        self.h = np.zeros([n_neurons + self.n_bias, batch_size])
        self.delta_ho = np.zeros([n_neurons, batch_size])
        self.grad_Phi = np.zeros([n_neurons, batch_size])

        # if a kernel mixture network is used and this is the last layer:
        # store kernel means and standard deviations
        if loss == 'kernel_mixture' and r == n_layers:
            self.kernel_means = kwargs['kernel_means']
            self.kernel_stds = kwargs['kernel_stds']

        # if parametric relu activation is used, set the value of a (x if x>0; a*x otherwise)
        if activation == 'parametric_relu':
            self.relu_a = kwargs['relu_a']

    def meet_the_neighbors(self, layer_rm1, layer_rp1):
        """
        Connect this layer to its neighbors

        Parameters
        ----------
        layer_rm1 : Layer object or None
            The layer before at index r - 1.
        layer_rp1 : Layer object or None
            The layer after at index r + 1.

        Returns
        -------
        None.

        """
        # if this layer is an input layer
        if self.r == 0:
            self.layer_rm1 = None
            self.layer_rp1 = layer_rp1
        # if this layer is an output layer
        elif self.r == self.n_layers:
            self.layer_rm1 = layer_rm1
            self.layer_rp1 = None
        # if this layer is hidden
        else:
            self.layer_rm1 = layer_rm1
            self.layer_rp1 = layer_rp1

        # fill the layer with neurons
        if self.r != 0:
            self.init_weights()

    def init_weights(self):
        """
        Initialize the weights and other related matrices of this layer

        Returns
        -------
        None.

        """
        # weights
        self.W = np.random.randn(self.layer_rm1.n_neurons + self.layer_rm1.n_bias,
                                 self.n_neurons) * np.sqrt(1.0 / self.layer_rm1.n_neurons)
        # loss gradient
        self.L_grad_W = np.zeros([self.layer_rm1.n_neurons + self.layer_rm1.n_bias, self.n_neurons])
        # momentum
        self.V = np.zeros([self.layer_rm1.n_neurons + self.layer_rm1.n_bias, self.n_neurons])
        # squared gradient
        self.A = np.zeros([self.layer_rm1.n_neurons + self.layer_rm1.n_bias, self.n_neurons])
        # L2 regularization
        self.Lamb = np.ones([self.layer_rm1.n_neurons + self.layer_rm1.n_bias,
                             self.n_neurons]) * self.lamb

        # do not apply regularization to the bias terms
        if self.bias:
            self.Lamb[-1, :] = 0.0

    def compute_output(self, batch_size):
        """
        Compute the output of the current layer in one shot using matrix -
        vector/matrix multiplication.

        Parameters
        ----------
        batch_size : int
            The batch size.

        Returns
        -------
        None.

        """

        a = np.dot(self.W.T, self.layer_rm1.h)

        # apply activation to a
        if self.activation == 'linear':
            self.h = a
        elif self.activation == 'sigmoid':
            self.h = 1.0/(1.0 - np.exp(-a))
        elif self.activation == 'relu':
            self.h = np.maximum(np.zeros([a.shape[0], a.shape[1]]), a)
        elif self.activation == 'leaky_relu':
            aa = np.copy(a)
            idx_lt0 = np.where(a <= 0.0)
            aa[idx_lt0[0], idx_lt0[1]] *= 0.01
            self.h = aa
        elif self.activation == 'parametric_relu':
            aa = np.copy(a)
            idx_lt0 = np.where(a <= 0.0)
            aa[idx_lt0[0], idx_lt0[1]] *= self.relu_a
            self.h = aa
        elif self.activation == 'softplus':
            self.h = np.log(1.0 + np.exp(a))
        elif self.activation == 'tanh':
            self.h = np.tanh(a)
        elif self.activation == 'hard_tanh':

            aa = np.copy(a)
            idx_gt1 = np.where(a >= 1.0)
            idx_ltm1 = np.where(a <= -1.0)
            aa[idx_gt1[0], idx_gt1[1]] = 1.0
            aa[idx_ltm1[0], idx_ltm1[1]] = -1.0

            self.h = aa

        else:
            print('Unknown activation type')
            sys.exit()

        # add bias neuron output
        if self.bias:
            self.h = np.vstack([self.h, np.ones(batch_size)])
        self.a = a

        # compute the gradient of the activation function,
        self.compute_grad_Phi()

    def compute_grad_Phi(self):
        """
        Compute the gradient in the activation function Phi wrt its input

        Returns
        -------
        None.

        """

        if self.activation == 'linear':
            self.grad_Phi = np.ones([self.n_neurons, self.batch_size])
        elif self.activation == 'sigmoid':
            self.grad_Phi = self.h[0:self.n_neurons]*(1.0 - self.h[0:self.n_neurons])
        elif self.activation == 'relu':
            idx_lt0 = np.where(self.a < 0.0)
            self.grad_Phi = np.ones([self.n_neurons, self.batch_size])
            self.grad_Phi[idx_lt0[0], idx_lt0[1]] = 0.0
        elif self.activation == 'leaky_relu':
            idx_lt0 = np.where(self.a < 0.0)
            self.grad_Phi = np.ones([self.n_neurons, self.batch_size])
            self.grad_Phi[idx_lt0[0], idx_lt0[1]] = 0.01
        elif self.activation == 'parametric_relu':
            idx_lt0 = np.where(self.a < 0.0)
            self.grad_Phi = np.ones([self.n_neurons, self.batch_size])
            self.grad_Phi[idx_lt0[0], idx_lt0[1]] = self.relu_a
        elif self.activation == 'softplus':
            self.grad_Phi = 1.0 / (1.0 + np.exp(-self.a))
        elif self.activation == 'tanh':
            self.grad_Phi = 1.0 - self.h[0:self.n_neurons]**2
        elif self.activation == 'hard_tanh':
            idx = np.where(np.logical_and(self.a > -1.0, self.a < 1.0))
            self.grad_Phi = np.zeros([self.n_neurons, self.batch_size])
            self.grad_Phi[idx[0], idx[1]] = 1.0

    def compute_loss(self, h, y_i):
        """
        Compute the value of the loss function.


        Parameters
        ----------
        h : array
            The activation of the output layer.
        y_i : array
            The target data.

        Returns
        -------
        None.

        """

        # only compute if in an output layer
        if self.layer_rp1 is None:
            if self.loss == 'perceptron_crit':
                self.L_i = np.max([-y_i * h, 0.0])
            elif self.loss == 'hinge':
                self.L_i = np.max([1.0 - y_i * h, 0.0])
            elif self.loss == 'logistic':
                self.L_i = np.log(1.0 + np.exp(-y_i * h))
            elif self.loss == 'squared':
                self.L_i = (y_i - h)**2
            elif self.loss == 'cross_entropy':
                # compute values of the softmax layer
                # more than 1 (independent) softmax layer can be placed at the output
                o_i = []
                [o_i.append(np.exp(h_i) / np.sum(np.exp(h_i), axis=0))
                 for h_i in np.split(h, self.n_softmax)]
                self.o_i = np.concatenate(o_i)
                # cross entropy loss with a softmax layer
                try:
                    self.L_i = -np.sum(y_i * np.log(self.o_i))
                except RuntimeWarning:
                    # if a neuron does not fire, e.g. in the case of RelU activation,
                    # we get a RuntimeWarning. Just add a small constant in this case.
                    self.L_i = -np.sum(y_i * np.log(self.o_i + 1e-20))
            elif self.loss == 'kernel_mixture' and self.n_softmax > 0:

                if y_i.ndim == 1:
                    y_i = y_i.reshape([1, y_i.size])

                self.L_i = 0.0
                idx = 0
                self.o_i = []
                self.p_i = []
                for h_i in np.split(h, self.n_softmax):
                    o_i = np.exp(h_i) / np.sum(np.exp(h_i), axis=0)
                    K_i = norm.pdf(y_i[idx], self.kernel_means[idx], self.kernel_stds[idx])
                    p_i = K_i * np.exp(h_i) / np.sum(K_i * np.exp(h_i), axis=0)
#                    p_i = K_i*o_i/np.sum(K_i*o_i, axis=0)
                    self.L_i = self.L_i - np.log(np.sum(o_i * K_i, axis=0))

                    self.o_i.append(o_i)
                    self.p_i.append(p_i)

                    idx += 1

                self.o_i = np.concatenate(self.o_i)
                self.p_i = np.concatenate(self.p_i)

            else:
                print('Cannot compute loss: unknown loss and/or activation function')
                sys.exit()

    def compute_delta_hy(self):
        """
        Compute the gradient of the network output wrt the activation functions
        of this layer.

        Returns
        -------
        None.

        """

        # if this layer is the output layer
        if self.layer_rp1 is None:
            self.delta_hy = np.ones([self.n_neurons, self.batch_size])
        else:
            # get the delta_ho values of the next layer (layer r+1)
            delta_hy_rp1 = self.layer_rp1.delta_hy

            # get the grad_Phi values of the next layer
            grad_Phi_rp1 = self.layer_rp1.grad_Phi

            # the weight matrix of the next layer
            W_rp1 = self.layer_rp1.W

            self.delta_hy = np.dot(W_rp1, delta_hy_rp1 * grad_Phi_rp1)[0:self.n_neurons, :]

    def compute_delta_oo(self, y_i):
        """
        Initialize the value of delta_ho at the output layer. This is the gradient
        of the loss function wrt the output of the neural network.

        Parameters
        ----------
        y_i : array
            The target data.

        Returns
        -------
        None.

        """

        # if the neuron is in the output layer, initialze delta_oo
        if self.layer_rp1 is None:

            # compute the loss function
            self.compute_loss(self.h, y_i)

            h = self.h

            # for binary classification
            if self.loss == 'logistic' and self.activation == 'linear':

                self.delta_ho = -y_i * np.exp(-y_i * h) / (1.0 + np.exp(-y_i * h))

            elif self.loss == 'squared':  # and self.activation == 'linear':

                self.delta_ho = -2.0 * (y_i - h)
                # grad_loss = elementwise_grad(self.test)
                # self.delta_ho = grad_loss(self.h, y_i)

            # for multinomial classification
            elif self.loss == 'cross_entropy':

                # (see eq. 3.22 of Aggarwal book)
                self.delta_ho = self.o_i - y_i

            elif self.loss == 'kernel_mixture' and self.n_softmax > 0:

                self.delta_ho = self.o_i - self.p_i

        else:
            print('Can only initialize delta_oo in output layer')
            sys.exit()

    def compute_delta_ho(self):
        """
        Compute the gradient of the loss function wrt the activation functions
        of this layer.

        Returns
        -------
        None.

        """
        # get the delta_ho values of the next layer (layer r+1)
        delta_ho_rp1 = self.layer_rp1.delta_ho

        # get the grad_Phi values of the next layer
        grad_Phi_rp1 = self.layer_rp1.grad_Phi

        # the weight matrix of the next layer
        W_rp1 = self.layer_rp1.W

        self.delta_ho = np.dot(W_rp1, delta_ho_rp1 * grad_Phi_rp1)[0:self.n_neurons, :]

    def compute_y_grad_W(self):
        """
        Compute the gradient of the network output wrt the weights of this layer.

        Returns
        -------
        None.

        """
        h_rm1 = self.layer_rm1.h
        delta_hy_grad_Phi = self.delta_hy * self.grad_Phi
        self.y_grad_W = np.dot(h_rm1, delta_hy_grad_Phi.T)

    def compute_L_grad_W(self):
        """
        Compute the gradient of the loss function wrt the weights of this layer.

        Returns
        -------
        None.

        """
        h_rm1 = self.layer_rm1.h
        delta_ho_grad_Phi = self.delta_ho * self.grad_Phi
        self.L_grad_W = np.dot(h_rm1, delta_ho_grad_Phi.T)

    def back_prop(self, y_i, jacobian=False):
        """
        Perform the backpropogation operations of the current layer.

        Parameters
        ----------
        y_i : array
            The target data.

        Returns
        -------
        None.

        """

        if self.r == self.n_layers:
            self.compute_delta_oo(y_i)
        else:
            self.compute_delta_ho()
        self.compute_L_grad_W()

        if jacobian:
            self.compute_delta_hy()
            if self.r > 0:
                self.compute_y_grad_W()



"""
==============================================================================
CLASS FOR A KERNEL MIXTURE NETWORK
------------------------------------------------------------------------------
Author: W. Edeling
Source: Luca Ambrogioni et al, The Kernel Mixture Network:
        A Nonparametric Method for Conditional Density Estimation of
        Continuous Random Variables, 2017.
==============================================================================
"""

from itertools import chain, product
import numpy as np
from scipy.stats import norm
from ..campaign import Campaign
import easysurrogate as es


class KMN_Surrogate(Campaign):

    def __init__(self, **kwargs):
        print('Creating Kernel Mixture Network Object')
        self.name = 'KMN Surrogate'
        # create a Feature_Engineering object
        self.feat_eng = es.methods.Feature_Engineering()

    ############################
    # START COMMON SUBROUTINES #
    ############################

    def train(self, feats, target, n_iter,
              kernel_means, kernel_stds, n_softmax,
              lags=None, local=False,
              test_frac=0.0,
              n_layers=2, n_neurons=100,
              activation='leaky_relu',
              batch_size=64, lamb=0.0, **kwargs):
        """
        Perform back propagation to train the QSN

        Parameters
        ----------
        feats : feature array, or list of different feature arrays
        target : the target data
        lags : list of time lags per feature
        local : apply the surrogate locally at each grid point.
        n_iter : number of back propagation iterations
        kernel_means : achor points of the kernels
        kernel_stds : standard deviations of the kernels
        test_frac : Fraction of the data to use for training.
                    The default is 0.0.
        n_layers : The number of layers in the neural network. Includes the
                   input layer and the hidden layers. The default is 2.
        n_neurons : The number of neurons per layer. The default is 100.
        activation : Type of activation function. The default is 'leaky_relu'.
        batch_size : Mini batch size. The default is 64.
        lamb : L2 regularization parameter. The default is 0.0.

        Returns
        -------
        None.

        """

        self.lags = lags
        self.local = local

        # number of softmax layers (one per output)
        self.n_softmax = n_softmax

        # create all possible combinations of the specified kernel means and std devs
        self.kernel_means = []
        self.kernel_stds = []
        for i in range(self.n_softmax):
            combi = np.array(list(chain(product(kernel_means[i], kernel_stds[i]))))
            self.kernel_means.append(combi[:, 0].reshape([-1, 1]))
            self.kernel_stds.append(combi[:, 1].reshape([-1, 1]))

        # size of a single softmax layer
        self.n_bins = self.kernel_means[0].size

        # prepare the training data
        X_train, y_train = self.feat_eng.get_training_data(feats, target, lags=lags, local=local,
                                                           test_frac=test_frac)

        # get the maximum lag that was specified
        self.max_lag = self.feat_eng.max_lag

        # number of output neurons
        n_out = self.n_bins * self.n_softmax

        # create the feed-forward QSN
        self.neural_net = es.methods.ANN(X=X_train, y=y_train,
                                         n_layers=n_layers, n_neurons=n_neurons,
                                         n_softmax=self.n_softmax, n_out=n_out,
                                         loss='kernel_mixture',
                                         activation=activation, batch_size=batch_size,
                                         lamb=lamb, decay_step=10**4, decay_rate=0.9,
                                         standardize_X=True, standardize_y=False,
                                         save=False,
                                         kernel_means=self.kernel_means,
                                         kernel_stds=self.kernel_stds)

        print('===============================')
        print('Training Kernel Mixture Network...')

        # train network for N_iter mini batches
        self.neural_net.train(n_iter, store_loss=True)
        self.set_data_stats()
        if lags is not None:
            self.feat_eng.initial_condition_feature_history(feats)
        # flatten the kernel properties into a single vector (size=#output neurons),
        # used in predict subroutine
        self.kernel_means_flat = np.concatenate(self.kernel_means)
        self.kernel_stds_flat = np.concatenate(self.kernel_stds)

    def predict(self, X):
        """
        Make a stochastic prediction of the output y conditional on the
        input features [X_t, X_{t-lag1}, X_{t-lag2}, ...]

        Parameters
        ----------
        X: the state at the current (time) step. If the KMN is conditioned on
        more than 1 (time-lagged) variable, X must be a list containing all
        variables at the current time step.

        Returns
        -------
        Stochastic prediction of the output y

        """
        # feat_eng._predict handles the preparation of the features and returns
        # self._feed_forward(X)
        return self.feat_eng._predict(X, self._feed_forward)

    def _feed_forward(self, feat):

        feat = (feat - self.feat_mean) / self.feat_std
        # o_i = the probability mass function at the output layer
        # max_idx = the bin index with the highest probability
        o_i, max_idx, _ = self.neural_net.get_softmax(feat.reshape([1, self.neural_net.n_in]))
        self.o_i = o_i
        self.max_idx = max_idx
        # return random sample from the conditional kernel density estimate
        # TODO: implement rvs from softmax layer
        return norm.rvs(self.kernel_means_flat[max_idx], self.kernel_stds_flat[max_idx]).flatten()

    def save_state(self):
        """
        Save the state of the KMN surrogate to a pickle file
        """
        state = self.__dict__
        super().save_state(state=state, name=self.name)

    def load_state(self):
        """
        Load the state of the KMN surrogate from file
        """
        super().load_state(name=self.name)

    ##########################
    # END COMMON SUBROUTINES #
    ##########################

    def set_data_stats(self):
        """
        If the data were standardized, this stores the mean and
        standard deviation of the features in the ANN_Surrogate object.
        """
        if hasattr(self.neural_net, 'X_mean'):
            self.feat_mean = self.neural_net.X_mean
            self.feat_std = self.neural_net.X_std
        else:
            self.feat_mean = 0.0
            self.feat_std = 1.0

        if hasattr(self.neural_net, 'y_mean'):
            self.output_mean = self.neural_net.y_mean
            self.output_std = self.neural_net.y_std
        else:
            self.output_mean = 0.0
            self.output_std = 1.0



"""
Class for an artificial neural network.
"""

import pickle
import tkinter as tk
from tkinter import filedialog
import numpy as np
# from scipy.stats import rv_discrete
# import h5py

from .Layer import Layer


class ANN:
    """
    Class for an artificial neural network.
    """

    def __init__(self, X, y, alpha=0.001, decay_rate=1.0, decay_step=10**5, beta1=0.9,
                 beta2=0.999, lamb=0.0, n_out=1, loss='squared', activation='tanh',
                 activation_out='linear', n_softmax=0, n_layers=2, n_neurons=16,
                 bias=True, batch_size=1, param_specific_learn_rate=True,
                 save=True, on_gpu=False, name='ANN',
                 standardize_X=True, standardize_y=True, **kwargs):
        """
        Initialize the Artificial Neural Network object.

        Parameters
        ----------
        X : array
            The input features.
        y : array
            The target data.
        alpha : float, optional
            The learning rate. The default is 0.001.
        decay_rate : float, optional
            Factor multiplying the decay rate every decay_step iterations.
            The default is 1.0.
        decay_step : int, optional
            The number of training iterations after which decay_rate is lowered.
            The default is 10**5.
        beta1 : float, optional
            Momentum parameter controlling the moving average of the loss gradient.
            Used for the parameter-specific learning rate. The default is 0.9.
        beta2 : float, optional
            Parameter controlling the moving average of the squared gradient.
            Used for the parameter-specific learning rate. The default is 0.999.
        lamb : float, optional
            L2 weight regularization parameter. The default is 0.0.
        n_out : int, optional
            The number of output neurons. The default is 1.
        loss : string, optional
            The name of the loss function. The default is 'squared'.
        activation : string, optional
            The name of the activation function of the hidden layers.
            The default is 'tanh'.
        activation_out : string, optional
            The name of the activation function of the output layer.
            The default is 'linear'.
        n_softmax : int, optional
            The number of softmax layers attached to the output. The default is 0.
        n_layers : int, optional
            The number of layers, not counting the input layer. The default is 2.
        n_neurons : int, optional
            The number of neurons per hidden layer. The default is 16.
        bias : boolean, optional
            Use a bias neuron. The default is True.
        batch_size : int, optional
            The size of the mini batch. The default is 1.
        param_specific_learn_rate : boolean, optional
            Use parameter-specific learing rate. The default is True.
        save : boolean, optional
            Save the neural network to a pickle file after training.
            The default is True.
        on_gpu : boolean, optional
            Train the neural network on a GPU using cupy. NOT IMPLEMENTED IN THIS VERSION.
            The default is False.
        name : string, optional
            The name of the neural network. The default is 'ANN'.
        standardize_X : boolean, optional
            Standardize the features. The default is True.
        standardize_y : boolean, optional
            Standardize the target data. The default is True.

        Returns
        -------
        None.

        """

        # the features
        self.X = X

        # number of training data points
        self.n_train = X.shape[0]

        # the training outputs
        self.y = y

        # number of input nodes
        try:
            self.n_in = X.shape[1]
        except IndexError:
            self.n_in = 1

        # #use either numpy or cupy via xp based on the on_gpu flag
        # global xp
        # if on_gpu == False:
        #     import numpy as xp
        # else:
        #     import cupy as xp

        # self.on_gpu = on_gpu
        self.on_gpu = False

        # standardize the training data
        if standardize_X:

            self.X_mean = np.mean(X, axis=0)
            self.X_std = np.std(X, axis=0)
            self.X = (X - self.X_mean) / self.X_std

        if standardize_y:
            self.y_mean = np.mean(y, axis=0)
            self.y_std = np.std(y, axis=0)
            self.y = (y - self.y_mean) / self.y_std
        self.standardize_X = standardize_X
        self.standardize_y = standardize_y

        # number of layers (hidden + output)
        self.n_layers = n_layers

        # number of neurons in a hidden layer
        self.n_neurons = n_neurons

        # number of output neurons
        self.n_out = n_out

        # use bias neurons
        self.bias = bias

        # loss function type
        self.loss = loss

        # training rate
        self.alpha = alpha

        # L2 regularization parameter
        self.lamb = lamb

        # the rate of decay and decay step for alpha
        self.decay_rate = decay_rate
        self.decay_step = decay_step

        # momentum parameter
        self.beta1 = beta1

        # squared gradient parameter
        self.beta2 = beta2

        # use parameter specific learning rate
        self.param_specific_learn_rate = param_specific_learn_rate

        # activation function of the hidden layers
        self.activation = activation

        # activation function of the output layer
        self.activation_out = activation_out

        # number of sofmax layers
        self.n_softmax = n_softmax

        # save the neural network after training
        self.save = save
        self.name = name

        # additional variables/dicts etc that must be stored in the ann object
        self.aux_vars = kwargs

        # size of the mini batch used in stochastic gradient descent
        self.batch_size = batch_size

        self.loss_vals = []

        self.init_network(**kwargs)

    def init_network(self, **kwargs):
        """
        Set up the network structure by creating the Layer objects and
        connecting them together.

        Returns
        -------
        None.

        """

        self.layers = []

        # add the input layer
        self.layers.append(Layer(self.n_in, 0, self.n_layers, 'linear',
                                 self.loss, self.bias, batch_size=self.batch_size,
                                 lamb=self.lamb, on_gpu=self.on_gpu))

        # add the hidden layers
        for r in range(1, self.n_layers):
            self.layers.append(Layer(self.n_neurons, r, self.n_layers, self.activation,
                                     self.loss, self.bias, batch_size=self.batch_size,
                                     lamb=self.lamb, on_gpu=self.on_gpu))

        # add the output layer
        self.layers.append(
            Layer(
                self.n_out,
                self.n_layers,
                self.n_layers,
                self.activation_out,
                self.loss,
                batch_size=self.batch_size,
                lamb=self.lamb,
                n_softmax=self.n_softmax,
                on_gpu=self.on_gpu,
                **kwargs))

        self.connect_layers()

        self.print_network_info()

    def connect_layers(self):
        """
         Connect each layer in the neural network with its neighbours

        Returns
        -------
        None.

        """

        self.layers[0].meet_the_neighbors(None, self.layers[1])
        self.layers[-1].meet_the_neighbors(self.layers[-2], None)

        for i in range(1, self.n_layers):
            self.layers[i].meet_the_neighbors(self.layers[i - 1], self.layers[i + 1])

    def feed_forward(self, X_i, batch_size=1):
        """
        Run the network forward.

        Parameters
        ----------
        X_i : array
            The feauture array, needs to have shape [batch size, number of features].
        batch_size : int, optional
            The bath size. The default is 1.

        Returns
        -------
        array
            The prediction of the neural network.

        """

        # set the features at the output of in the input layer
        if not self.bias:
            self.layers[0].h = X_i.T
        else:
            self.layers[0].h = np.ones([self.n_in + 1, batch_size])
            self.layers[0].h[0:self.n_in, :] = X_i.T

        for i in range(1, self.n_layers + 1):
            # compute the output on the layer using matrix-maxtrix multiplication
            self.layers[i].compute_output(batch_size)

        return self.layers[-1].h

    def get_softmax(self, X_i):
        """
        Get the output of the softmax layer.

        Parameters
        ----------
        X_i : array
            The input features.

        Returns
        -------
        probs : array
            the probabilities of the softmax layer.
        idx_max : int
            The softmax output with the highest probability.

        """
        # feed forward features X_i
        h = self.feed_forward(X_i, batch_size=1)

        probs = []
        idx_max = []
        # rvs = []

        # split the output of the last layer over the number of softmax layers
        for h_i in np.split(h, self.n_softmax):
            # compute the softmax probabilities.
            o_i = np.exp(h_i) / np.sum(np.exp(h_i), axis=0)
            o_i = o_i / np.sum(o_i)
            probs.append(o_i)
            # the softmax output w
            idx_max.append(np.argmax(o_i))

            # draw a random sample from the discrete distribution o_i.
            # TODO: Slow for some reason. Find faster implementation.
            # pmf = rv_discrete(values=(np.arange(o_i.size), o_i.flatten()))
            # rvs.append(pmf.rvs())

        # return values and index of highest probability and random samples from pmf
        return probs, idx_max, None

    def jacobian(self, X_i, batch_size=1, feed_forward=True):
        """
        Compute the Jacobian of the neural net via back propagation.

        Parameters
        ----------
        X_i : array
            The input features.
        batch_size : int, optional
            The batch size. The default is 1.

        Returns
        -------
        array
            The Jacobian.

        """
        if feed_forward:
            self.feed_forward(X_i, batch_size=batch_size)

        for i in range(self.n_layers, -1, -1):
            self.layers[i].compute_delta_hy()
            # also compute the gradient of the output wrt the weights
            if i > 0:
                self.layers[i].compute_y_grad_W()

        # delta_hy of the input layer = the Jacobian of the neural net
        return self.layers[0].delta_hy

    def back_prop(self, y_i, jacobian=False):
        """
        Back-propagation algorithm to find gradient of the loss function with respect
        to the weights of the neural network.

        Parameters
        ----------
        y_i : array
            The target data on which to evaluate the loss funcion.

        Returns
        -------
        None.

        """

        # start back propagation over hidden layers, starting with output layer
        for i in range(self.n_layers, 0, -1):
            self.layers[i].back_prop(y_i, jacobian=jacobian)

    def batch(self, X_i, y_i, alpha=0.001, beta1=0.9, beta2=0.999):
        """
        Update the weights using a mini batch.

        Parameters
        ----------
        X_i : array
            The input features of the mini batch.
        y_i : array
            The target data of the mini batch.
        alpha : float, optional
            The learning rate. The default is 0.001.
        beta1 : float, optional
            Momentum parameter controlling the moving average of the loss gradient.
            Used for the parameter-specific learning rate. The default is 0.9.
        beta2 : float, optional
            Parameter controlling the moving average of the squared gradient.
            Used for the parameter-specific learning rate. The default is 0.999.

        Returns
        -------
        None.

        """
        self.feed_forward(X_i, self.batch_size)
        self.back_prop(y_i)

        for r in range(1, self.n_layers + 1):

            layer_r = self.layers[r]

            # momentum
            layer_r.V = beta1 * layer_r.V + (1.0 - beta1) * layer_r.L_grad_W
            # moving average of squared gradient magnitude
            layer_r.A = beta2 * layer_r.A + (1.0 - beta2) * layer_r.L_grad_W**2

            # select learning rate
            if not self.param_specific_learn_rate:
                # same alpha for all weights
                alpha_i = alpha
            # param specific learning rate
            else:
                # RMSProp
                alpha_i = alpha / (np.sqrt(layer_r.A + 1e-8))

                # Adam
                #alpha_t = alpha*np.sqrt(1.0 - beta2**t)/(1.0 - beta1**t)
                #alpha_i = alpha_t/(np.sqrt(layer_r.A + 1e-8))

            # gradient descent update step with L2 regularization
            if self.lamb > 0.0:
                layer_r.W = (1.0 - layer_r.Lamb * alpha_i) * layer_r.W - alpha_i * layer_r.V
            # without regularization
            else:
                layer_r.W = layer_r.W - alpha_i * layer_r.V

                # Nesterov momentum
                # layer_r.W += -alpha*beta1*layer_r.V

    # train the neural network
    def train(
            self,
            n_batch,
            store_loss=False,
            sequential=False,
            verbose=True):
        """
        Train the neural network using stochastic gradient descent.

        Parameters
        ----------
        n_batch : int
            The number of mini-batch iterations.
        store_loss : boolean, optional
            Store the values of the loss function. The default is False.
        sequential : boolean, optional
            Sample a sequential slab of data, starting from a random point.
            The default is False.
        verbose : boolean, optional
            Print information to screen while training. The default is True.

        Returns
        -------
        None.

        """

        for i in range(n_batch):

            # select a random training instance (X, y)
            if not sequential:
                rand_idx = np.random.randint(0, self.n_train, self.batch_size)
            # select a random starting point, and use sequential data from there
            else:
                if self.n_train > self.batch_size:
                    start = np.random.randint(0, self.n_train - self.batch_size, 1)
                else:
                    start = 0
                rand_idx = np.arange(start, start + self.batch_size)

            # compute learning rate
            alpha = self.alpha * self.decay_rate**(np.int(i / self.decay_step))

            # run the batch
            self.batch(
                self.X[rand_idx],
                self.y[rand_idx].T,
                alpha=alpha,
                beta1=self.beta1,
                beta2=self.beta2)

            # store the loss value
            if store_loss:
                # l = 0.0
                # for k in range(self.n_out):
                    # l += self.layers[-1].L_i
                
                l = self.layers[-1].L_i

                if np.mod(i, 1000) == 0:
                    loss_i = np.mean(l)
                    if verbose:
                        print('Batch', i, 'learning rate', alpha, 'loss:', loss_i)
                    # note: appending a cupy value to a list is inefficient -
                    # if done every iteration
                    # it will slow down executing significantly
                    self.loss_vals.append(loss_i)

        if self.save:
            self.save_ANN()

    def save_ANN(self, file_path="", store_data=False):
        """
        Save the neural network to a picke file.

        Parameters
        ----------
        file_path : string, optional
            The full path of the pickle file. The default is "", in which case a
            filedialog window opens.
        store_data : boolean, optional
            Store the training data as well. The default is False, since storing a
            large amount of data in a pickel file is inefficient.

        Returns
        -------
        None.

        """
        if len(file_path) == 0:

            root = tk.Tk()
            root.withdraw()

            file = filedialog.asksaveasfile(title="Save network",
                                            mode='wb', defaultextension=".pickle")
        else:
            file = open(file_path, 'wb')

        print('Saving ANN to', file.name)

        if store_data:
            # store everything, also data
            pickle.dump(self.__dict__, file)
        else:
            tmp = self.__dict__.copy()
            # do not store data to pickle
            tmp['X'] = []
            tmp['y'] = []
            pickle.dump(tmp, file)

        file.close()

    def load_ANN(self, file_path=""):
        """
        Load the neural network from a pickle file.

        Parameters
        ----------
        file_path : string, optional
            The full path of the pickle file. The default is "", in which case a
            filedialog window opens.

        Returns
        -------
        None.

        """

        # select file via GUI is file_path is not specified
        if len(file_path) == 0:

            root = tk.Tk()
            root.withdraw()

            file_path = filedialog.askopenfilename(title="Open neural network",
                                                   filetypes=(('pickle files', '*.pickle'),
                                                              ('All files', '*.*')))

        print('Loading ANN from', file_path)

        file = open(file_path, 'rb')
        self.__dict__ = pickle.load(file)
        file.close()

        if self.__dict__['X'] == []:
            print('===============================')
            print('**Warning: ANN was saved without training data**')
            print('===============================')

        self.print_network_info()

    def set_batch_size(self, batch_size):
        """
        Set the batch size in all the layers.

        Parameters
        ----------
        batch_size : int
            The batch size.

        Returns
        -------
        None.

        """

        self.batch_size = batch_size

        for i in range(self.n_layers + 1):
            self.layers[i].batch_size = batch_size

    def compute_misclass_softmax(self, X=None, y=None):
        """
        Compute the number of misclassifications for the sofmax layer(s).

        Parameters
        ----------
        X : array, optional
            Feature array. The default is None, in which case the entire training set is used.
        y : array, optional
            Target data array. The default is None, in which case the entire
            training set is used.

        Returns
        -------
        float
            The misclassification percentage in [0,1] per softmax layer.

        """

        n_misclass = np.zeros(self.n_softmax)

        # compute misclassification error of the training set if X and y are not set
        if y is None:
            print('Computing number of misclassifications wrt all training data.')
            X = self.X
            y = self.y
        else:
            print('Computing number of misclassifications wrt specified data (',
                  X.shape[0], 'samples)')

        n_samples = X.shape[0]

        for i in range(n_samples):
            _, max_idx_ann, _ = self.get_softmax(X[i].reshape([1, self.n_in]))

            max_idx_data = np.array([np.where(y_j == 1.0)[0]
                                     for y_j in np.split(y[i], self.n_softmax)])

            for j in range(self.n_softmax):
                if max_idx_ann[j] != max_idx_data[j]:
                    n_misclass[j] += 1

            if np.mod(i, 1000) == 0:
                print('Computing misclassification error:',
                      np.around(i / n_samples * 100, 1), '%')

        print('Number of misclassifications =', n_misclass)
        print('Misclassification percentage =', n_misclass / n_samples * 100, '%')

        return n_misclass / n_samples

    def get_n_weights(self):
        """
        Return the number of weights

        Returns
        -------
        n_weights : int
            The number of weights.

        """

        n_weights = 0

        for i in range(1, self.n_layers + 1):
            n_weights += self.layers[i].W.size

        print('This neural network has', n_weights, 'weights.')

        return n_weights

    def print_network_info(self):
        """
        Print some characteristics of the neural network to screen.

        Returns
        -------
        None.

        """
        print('===============================')
        print('Neural net parameters')
        print('===============================')
        print('Number of layers =', self.n_layers)
        print('Number of features =', self.n_in)
        print('Loss function =', self.loss)
        print('Number of neurons per hidden layer =', self.n_neurons)
        print('Number of output neurons =', self.n_out)
        print('Activation hidden layers =', self.activation)
        print('Activation output layer =', self.activation_out)
        # print('On GPU =', self.on_gpu)
        self.get_n_weights()
        print('===============================')



"""
===============================================================================
CLASS FOR A REDUCED SURROGATE FOR SPECTRAL MODELS
------------------------------------------------------------------------------
Author: W. Edeling
Source: W. Edeling, D. Crommelin, Reducing data-driven dynamical subgrid scale
models by physical constraints, Computer & Fluids, 2020.
===============================================================================
"""

import numpy as np
import easysurrogate as es
from ..campaign import Campaign


class Reduced_Surrogate(Campaign):
    """
    Reduced Surrogate class
    """

    def __init__(self, n_qoi, n_model_1d):
        """
        Create a Reduced Surrogate object.

        Parameters
        ----------
        n_qoi : int
            The number of integrated QoI to track.
        n_model_1d : int
            The number of grid points in one direction.

        Returns
        -------
        None.

        """
        self.n_qoi = n_qoi
        self.n_model_1d = n_model_1d
        print('Creating Reduced Surrogate Object')
        self.name = 'Reduced Surrogate'

    ############################
    # START COMMON SUBROUTINES #
    ############################

    def train(self, V, dQ):
        """
        A training step of the reduced surrogate.

        Parameters
        ----------
        V : list
            A list containing the Fourier coefficients of reduced basis functions V_i.
        dQ : array
            The reference statistics at a given time step - the same statistics computed
            with the lower resolution model.

        Returns
        -------
        dict
            A dictionary containing:
            'sgs_hat': the reduced subgrid-scale term,
            'c_ij': the coefficients of the linear system A_i * c_i = b_i (Eq 20-21 in the paper),
            'inner_prods': the inner products (V_i, T_{i,j}) (Eq 21),
            'src_Q': the inner product (V_i, P_i) that is part of a source term of Q (Eq 23),
            'tau': the multiplier of (V_i, P_i) (Eq 23).

        """

        assert isinstance(dQ, np.ndarray), 'dQ must be of type numpy.ndarray'

        assert dQ.size == self.n_qoi, 'size dQ_data must be %d' % self.n_qoi

        V_hat = np.zeros([self.n_qoi, self.n_model_1d, self.n_model_1d]) + 0.0j
        for i in range(self.n_qoi):
            V_hat[i] = V[i]

        return self.reduced_r(V_hat, dQ)

    def generate_online_training_data(self, feats, LR_before, LR_after, HR_before, HR_after,
                                      qoi_func, **kwargs):
        """
        Compute the features and the target data for an online training step. Results are
        stored internally, and used within the 'train_online' subroutine.

        Source:
        Rasp, "Coupled online learning as a way to tackle instabilities and biases
        in neural network parameterizations: general algorithms and Lorenz 96
        case study", 2020.

        Parameters
        ----------
        feats : array or list of arrays
            The input features
        LR_before : array or list or arrays
            Low resolution state(s) at previous time step.
        LR_after : array or list or arrays
            Low resolution state(s) at current time step.
        HR_before : array or list or arrays
            High resolution state(s) at previous time step.
        HR_after : array or list or arrays
            High resolution state(s) at current time step.
        qoi_func : function
            A user-specfied function f(state, **kwargs) what computes the QoI from the LR
            or HR state.

        Returns
        -------
        None.

        """

        # multiple features arrays are stored in a list. For consistency put a single
        # array also in a list.
        if isinstance(feats, np.ndarray):
            feats = [feats]

        # allow the state to be stored in a list in case there are multiple states
        if isinstance(LR_before, np.ndarray):
            LR_before = [LR_before]
        if isinstance(LR_after, np.ndarray):
            LR_after = [LR_after]
        if isinstance(HR_before, np.ndarray):
            HR_before = [HR_before]
        if isinstance(HR_after, np.ndarray):
            HR_after = [HR_after]

        # store input features
        for i in range(self.n_feat_arrays):
            self.dQ_surr.feat_eng.online_feats[i].append(feats[i])

        dQ = []
        n_HR = HR_before[0].shape[0]
        # loop over all states
        for i in range(len(LR_before)):

            # project the low-res model to the high-res grid
            LR_before_projected = self.up_scale(LR_before[i], n_HR)

            # the difference between the low res and high res model (projected to low-res grid)
            # at time n
            delta_nudge = LR_before_projected - HR_before[i]

            # the estimated state of the (projected) HR model would there have been no nudging
            HR_no_nudge = HR_after[i] - delta_nudge / self.tau_nudge * self.dt_LR

            # compute the HR QoI
            Q_HR = qoi_func(HR_no_nudge, **kwargs)
            # Q_HR = qoi_func(HR_after[i], **kwargs)

            # compute the LR QoI
            # Note: could store multiple functions in a list if required.
            Q_LR = qoi_func(LR_after[i], **kwargs)

            dQ.append(Q_HR - Q_LR)

        # The difference in HR and LR QoI is the target of the surrogate
        self.dQ_surr.feat_eng.online_target.append(np.concatenate(dQ))

        # remove oldest item from the online features is the window lendth is exceeded
        if len(self.dQ_surr.feat_eng.online_feats[0]) > self.window_length:
            for i in range(self.n_feat_arrays):
                self.dQ_surr.feat_eng.online_feats[i].pop(0)
            self.dQ_surr.feat_eng.online_target.pop(0)

    def set_online_training_parameters(self, tau_nudge, dt_LR, window_length):
        """
        Stores parameters required for online training.

        Parameters
        ----------
        tau_nudge : float
            Nudging time scale.
        dt_LR : float
            Time step low resolution model.
        window_length : int
            The length of the moving window in which online features are stored.

        Returns
        -------
        None.

        """
        assert hasattr(self, 'dQ_surr'), "A surrogate for dQ must be set using set_dQ_surr(...)"
        # store the parameters in the Feature Engineering object of the dQ surrogate
        self.dQ_surr.feat_eng.set_online_training_parameters(tau_nudge, dt_LR, window_length)
        # als store it locally
        self.tau_nudge = tau_nudge
        self.dt_LR = dt_LR
        self.window_length = window_length
        # the number of feature vectors for the dQ surrogate
        self.n_feat_arrays = self.dQ_surr.feat_eng.n_feat_arrays

    def predict(self, V, dQ_surr):
        """
        Compute the reduced surrogate subgrid-scale term using a surrogate for dQ := qoi_ref -
        qoi_model

        Parameters
        ----------
        V : list
            A list containing the Fourier coefficients of reduced basis functions V_i.
        dQ_surr : array
            A surrogate prediction of dQ.

        Returns
        -------
        dict
            A dictionary containing:
            'sgs_hat': the reduced subgrid-scale term,
            'c_ij': the coefficients of the linear system A_i * c_i = b_i (Eq 20-21 in the paper),
            'inner_prods': the inner products (V_i, T_{i,j}) (Eq 21),
            'src_Q': the inner product (V_i, P_i) that is part of a source term of Q (Eq 23),
            'tau': the multiplier of (V_i, P_i) (Eq 23).

        """
        # the code is acutally the same as in the training step, only dQ is computed using a
        # surrogate, instead of extracted from the high-resolution model.
        return self.train(V, dQ_surr)

    def save_state(self):
        """
        Save the state of the reduced surrogate to a pickle file.

        Returns
        -------
        None.

        """

        state = self.__dict__
        super().save_state(state=state, name=self.name)

    def load_state(self):
        """
        Load the state of the reduced surrogate from file.

        Returns
        -------
        None.

        """
        super().load_state(name=self.name)

    ##########################
    # END COMMON SUBROUTINES #
    ##########################

    def set_dQ_surr(self, surrogate):
        """
        Set the surrogate used for predicting dQ := Q_reference - Q_model

        Parameters
        ----------
        surrogate : object
            A surrogate trained on dQ data.

        Returns
        -------
        None.

        """
        self.dQ_surr = surrogate

    def reduced_r(self, V_hat, dQ):
        """
        Compute the reduced SGS term

        Parameters
        ----------
        V_hat : list
            A list containing the Fourier coefficients of reduced basis functions V_i.
        dQ : array
            The difference in the QoI between the high and low resolution model.

        Returns
        -------
        reduced_dict : dict
            A dictionary containing:
            'sgs_hat': the reduced subgrid-scale term,
            'c_ij': the coefficients of the linear system A_i * c_i = b_i (Eq 20-21 in the paper),
            'inner_prods': the inner products (V_i, T_{i,j}) (Eq 21),
            'src_Q': the inner product (V_i, P_i) that is part of a source term of Q (Eq 23),
            'tau': the multiplier of (V_i, P_i) (Eq 23).
        """

        # compute the T_ij basis functions
        T_hat = np.zeros([self.n_qoi, self.n_qoi, self.n_model_1d, self.n_model_1d]) + 0.0j

        for i in range(self.n_qoi):

            T_hat[i, 0] = V_hat[i]

            J = np.delete(np.arange(self.n_qoi), i)

            idx = 1
            for j in J:
                T_hat[i, idx] = V_hat[j]
                idx += 1

        # compute the coefficients c_ij
        inner_prods = self.inner_products(V_hat)

        c_ij = self.compute_cij_using_V_hat(inner_prods)

        sgs_hat = 0.0

        src_Q = np.zeros(self.n_qoi)
        tau = np.zeros(self.n_qoi)

        # loop over all QoI
        for i in range(self.n_qoi):
            # compute the fourier coefs of the P_i
            P_hat_i = T_hat[i, 0]
            for j in range(0, self.n_qoi - 1):
                P_hat_i -= c_ij[i, j] * T_hat[i, j + 1]

            # (V_i, P_i) integral
            src_Q_i = compute_int(V_hat[i], P_hat_i, self.n_model_1d)

            # compute tau_i = Delta Q_i/ (V_i, P_i)
            tau_i = dQ[i] / src_Q_i

            src_Q[i] = src_Q_i
            tau[i] = tau_i

            # compute reduced subgrid-scale source term
            sgs_hat -= tau_i * P_hat_i

        reduced_dict = {'sgs_hat': sgs_hat, 'c_ij': c_ij,
                        'inner_prods': np.triu(inner_prods),
                        'src_Q': src_Q, 'tau': tau}

        return reduced_dict

    def compute_cij_using_V_hat(self, inner_prods):
        """
        Compute the coefficients c_ij of P_i = T_{i,1} - c_{i,2}*T_{i,2}, - ... (Eq 16, 20 and 21)

        Parameters
        ----------
        inner_prods : array
            A matrix containing all inner products (V_i, T_{i,j}).

        Returns
        -------
        c_ij : array
            The coefficients of the subgrid-scale basis functions (Eq 16, 20 and 21) at the
            current time step.

        """
        c_ij = np.zeros([self.n_qoi, self.n_qoi - 1])

        for i in range(self.n_qoi):
            A = np.zeros([self.n_qoi - 1, self.n_qoi - 1])
            b = np.zeros(self.n_qoi - 1)

            k = np.delete(np.arange(self.n_qoi), i)

            for j1 in range(self.n_qoi - 1):
                for j2 in range(j1, self.n_qoi - 1):
                    A[j1, j2] = inner_prods[k[j1], k[j2]]
                    if j1 != j2:
                        A[j2, j1] = A[j1, j2]

            for j1 in range(self.n_qoi - 1):
                b[j1] = inner_prods[i, k[j1]]

            if self.n_qoi == 2:
                c_ij[i, :] = b / A
            else:
                c_ij[i, :] = np.linalg.solve(A, b)

        return c_ij

    def inner_products(self, V_hat):
        """
        Compute all the inner products (V_i, T_{i,j})

        Parameters
        ----------
        V_hat : list
            A list containing the Fourier coefficients of reduced basis functions V_i.

        Returns
        -------
        inner_prods : array
            A matrix containing all inner products (V_i, T_{i,j}).

        """
        V_hat = V_hat.reshape([self.n_qoi, self.n_model_1d**2])

        return np.dot(V_hat, np.conjugate(V_hat).T).real / self.n_model_1d**4

    def down_scale(self, X_hat, N_LR):
        """
        Down-scale X to a lower spatial resolution by removing high-frequency Fourier coefficients.

        Parameters
        ----------
        X_hat : array (complex)
            An array of Fourier coefficients of X, where the spatial resolution in 1D is higher than N.
        N_LR : int
            The new, lower, spatial resolution (X.shape[0] < N).

        Returns
        -------
        X_hat : array (complex)
            The Fourier coefficients of X at a spatial resolution determined by N.

        """
        assert N_LR < X_hat.shape[0], "N must be smaller than X_hat.shape[0] to down scale."

        # The spatial dimension of the problem
        d = X_hat.ndim
        # The HR grid resolution
        N_HR = X_hat.shape[0]

        # the range that should be deleted
        start = int(N_LR / 2)
        end = X_hat.shape[0] - start
        # Remove the Fourier coefficients that are not present in the lower-resolution version
        # of X_hat
        for i in range(X_hat.ndim):
            X_hat = np.delete(X_hat, np.arange(start, end), axis=i)
        # In numpy only the inverse transform is scaled. The following term must be applied
        # to ensure the correct scaling factor is applied in the inverse transform.
        scaling_factor = (N_LR / N_HR)**d
        return X_hat * scaling_factor

    def up_scale(self, X_hat, N_HR):
        """
        Up-scale X to a higher spatial resolution by padding high-frequency Fourier coefficients with
        zeros. Thus far this will only work for 1 or 2 dimensional arrays.


        Parameters
        ----------
        X_hat : array (complex)
            The Fourier coefficients of X.
        N_HR : int
            The new, higher, spatial resolution (X.shape[0] > N).

        Returns
        -------
        X_hat : array (complex)
            The Fourier coefficients of X at a higher spatial resolution determined ny N.

        """

        N_LR = X_hat.shape[0]
        d = X_hat.ndim

        # assert N_LR < N_HR, "X_hat.shape[0] must be < N_HR in order to upscale X_hat."
        assert d == 1 or d == 2, "Upscaling only implemented for 1d or 2d arrays."

        start = int(N_LR / 2)
        pad_size = N_HR - N_LR
        if X_hat.ndim == 1:
            # pad the 1d array with zeros
            X_hat = np.insert(X_hat, start, np.zeros(pad_size) + 0j)
        elif X_hat.ndim == 2:
            # pad the 2d array with a 'cross' of zeros
            X_hat = np.insert(X_hat, start, np.zeros([pad_size, N_LR]) + 0j, axis=0)
            X_hat = np.insert(X_hat, start, np.zeros([pad_size, N_HR]) + 0j, axis=1)
        # In numpy only the inverse transform is scaled. The following term must be applied
        # to ensure the correct scaling factor is applied in the inverse transform.
        scaling_factor = (N_HR / N_LR)**d
        return X_hat * scaling_factor


def compute_int(X1_hat, X2_hat, N):
    """
    Compute the integral of X1*X2 over the spatial domain using the Fourier expansion,
    see section 4.4 of the paper.

    Parameters
    ----------
    X1_hat : array (complex)
        The Fourier coefficients of X1.
    X2_hat : array (complex)
        The Fourier coefficients of X2.
    N : int
        The number of points in one spatial direction.

    Returns
    -------
    float
        The value of the integral.

    """
    integral = np.dot(X1_hat.flatten(), np.conjugate(X2_hat.flatten())) / N**4
    return integral.real



"""
Cumulative sum layer
"""
import numpy as np

from .Layer import Layer

class CumSum_Layer(Layer):
    
    def __init__(self, n_neurons, r, n_layers, loss, batch_size):
        """
        Initialize a CumSum layer object.      

        Parameters
        ----------
        n_neurons : int
            The number of neurons. Must match the number of neurons of the output layer..
        r : int
            The layer index.
        n_layers : int
            The number of layers in the network.
        loss : string
            The loss function.
        batch_size : int
            The batch size.

        Returns
        -------
        None.

        """
        super().__init__(n_neurons, r, n_layers, 'linear', loss, 
                         bias=False, batch_size=batch_size)
    
    def compute_output(self, batch_size):
        self.h = np.cumsum(self.layer_rm1.h, axis=0)
        # compute the gradient of the activation function,
        self.compute_grad_Phi()

    def init_weights(self):
        """
        Initialize the weights and other related matrices of this layer. As this
        layer will be attached to the output, we assume the previous layers does
        not have bias neurons.

        Returns
        -------
        None.

        """
        # weights corresponding to a cumulative sum.
        self.W = np.triu(np.ones([self.layer_rm1.n_neurons, self.n_neurons]))
        # loss gradient
        self.L_grad_W = np.zeros([self.layer_rm1.n_neurons, self.n_neurons])
        # momentum
        self.V = np.zeros([self.layer_rm1.n_neurons, self.n_neurons])
        # squared gradient
        self.A = np.zeros([self.layer_rm1.n_neurons, self.n_neurons])
        # L2 regularization
        self.Lamb = np.ones([self.layer_rm1.n_neurons, self.n_neurons]) * self.lamb
        
    def compute_L_grad_W(self):
        """
        Compute the gradient of the loss function wrt the weights of this layer.

        Returns
        -------
        None.

        """
        return self.L_grad_W
    


"""
CLASS TO PERFORM ANALYSIS ON RESULTS FROM A QUANTIZED SOFTMAX SURROGATE.
"""
from .base import BaseAnalysis
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import animation


class QSN_analysis(BaseAnalysis):
    """
    QSN analysis class
    """

    def __init__(self, qsn_surrogate, **kwargs):
        print('Creating QSN_analysis object')
        self.qsn_surrogate = qsn_surrogate

    def get_classification_error(self, index=None, **kwargs):
        """
        Compute the misclassification error of the QSN surrogate.

        Parameters
        ----------

        index : array, default is None
            indices to select a subset of feature/data points to perform test on. When None
            the classification error is computed on the entire dataset.

        Returns
        -------
        Prints classification error to screen for every softmax layer

        """

        if index is not None:
            X = self.qsn_surrogate.neural_net.X[index]
            y = self.qsn_surrogate.neural_net.y[index]
        else:
            X = self.qsn_surrogate.neural_net.X
            y = self.qsn_surrogate.neural_net.y

        self.qsn_surrogate.neural_net.compute_misclass_softmax(X=X, y=y)

    # def make_movie(self, n_frames=500):
    #     """
    #     Makes a move using the training data. Left subplot shows the evolution of the
    #     kernel-density estimate, and the right subplot show the time series of the data
    #     and the random samples drawm from the kernel-density estimate. Saves the movie
    #     to a .gif file.

    #     Parameters

    #     n_frames (int): default is 500
    #         The number of frames to use in the movie.

    #     Returns: None
    #     """

    #     # get the (normalized, time-lagged) training data from the neural network
    #     X = self.qsn_surrogate.neural_net.X
    #     y = self.qsn_surrogate.neural_net.y

    #     print('===============================')
    #     print('Making movie...')

    #     # list to store the movie frames in
    #     ims = []
    #     fig = plt.figure(figsize=[8, 4])
    #     ax1 = fig.add_subplot(121, xlabel=r'$B_k$', ylabel=r'', yticks=[])
    #     ax2 = fig.add_subplot(122, xlabel=r'time', ylabel=r'$B_k$')
    #     plt.tight_layout()

    #     # number of features
    #     n_feat = X.shape[1]
    #     # number of softmax layers
    #     n_softmax = self.qsn_surrogate.n_softmax

    #     # allocate memory
    #     samples = np.zeros([n_frames, n_softmax])

    #     # make movie by evaluating the network at TRAINING inputs
    #     for i in range(n_frames):

    #         # draw a random sample from the network
    #         o_i, idx_max, _ = self.qsn_surrogate.neural_net.get_softmax(X[i].reshape([1, n_feat]))
    #         self.qsn_surrogate.sampler.resample(idx_max)
    #         if np.mod(i, 100) == 0:
    #             print('i =', i, 'of', n_frames)

    #         # create a single frame, store in 'ims'
    #         plt2 = ax1.plot(range(self.qsn_surrogate.n_bins),
    #                         np.zeros(self.qsn_surrogate.n_bins),
    #                         o_i[0], 'b', label=r'conditional pmf')
    #         plt3 = ax1.plot(y[i][0], 0.0, 'ro', label=r'data')
    #         plt4 = ax2.plot(y[0:i, 0], 'ro', label=r'data')
    #         plt5 = ax2.plot(samples[0:i, 0], 'g', label='random sample')

    #         if i == 0:
    #             ax1.legend(loc=1, fontsize=9)
    #             ax2.legend(loc=1, fontsize=9)

    #         ims.append((plt2[0], plt3[0], plt4[0], plt5[0],))

    #     # make a movie of all frame in 'ims'
    #     im_ani = animation.ArtistAnimation(fig, ims, interval=20,
    #                                        repeat_delay=2000, blit=True)
    #     im_ani.save('kvm.gif')
    #     print('done. Saved movie to qsn.gif')



"""
===============================================================================
CLASS FOR FEATURE ENGINEERING SUBROUTINES
===============================================================================
"""

import numpy as np
from itertools import chain
from scipy import stats


class Feature_Engineering:
    """
    Feature_Engineering class. containing several generic methods for the manipulation of features
    that are shared by all surrogate methods.
    """

    def __init__(self):
        print('Creating Feature Engineering object')
        self.lags = None
        self.local = False

    def _predict(self, X, feed_forward):
        """
        Contains the generic processing of features that is independent of the chosen surrogate
        method. Features are processined depending upon the presence of time lags or the local /
        non-local nature of the surroagte. The processed features are then passed to the
        surrogate-specific feed_forward method.

        Parameters
        ----------
        X : array or list of arrays
            The feature array or list of feature arrays on which to evaluate the surrogate.
        feed_forward : function
            The prediction function that is specific to a particular surrogate.

        Returns
        -------
        array
        feed_forward(X)

        """

        if not isinstance(X, list):
            X = [X]

        # make sure all feature vectors have the same ndim.
        # This will raise an error when for instance X1.shape = (10,) and X2.shape = (10, 1)
        ndims = [X_i.ndim for X_i in X]
        assert all([ndim == ndims[0] for ndim in ndims]), "All features must have the same ndim"

        # make sure features are at most two dimensional arrays
        assert ndims[0] <= 2, "Only 1 or 2 dimensional arrays are allowed as features."

        # in the case of two dimensional arrays, make are the second dimension is the same
        # for all features. This dimension must equal the grid size
        local = False
        if ndims[0] == 2:
            local = True
            shapes1 = [X_i.shape[1] for X_i in X]
            assert all([shape1 == shapes1[0] for shape1 in shapes1]), \
                "The size of the second dimension must be the same for all features."
            # if a second dimension is specified, it is assumed that we must loop over this
            # dimension in order to make a single prediction
            n_points = shapes1[0]

        # time-lagged surrogate
        if self.lags is not None:

            # append the current state X to the feature history
            self.append_feat(X)

            # if not local, get entire feature vector and feed forward
            if not local:
                feat = self.get_feat_history()
                return feed_forward(feat)
            # if local, loop over the 2nd dimension of the feature vector and feed forward
            # every entry
            else:
                y = []
                for p in range(n_points):
                    feat = self.get_feat_history(index=p)
                    y.append(feed_forward(feat))
                return np.array(y).flatten()
        # no lags
        else:
            # no lags and non local, create a single vector of X and feed forward
            if not local:
                feat = np.concatenate(X)
                return feed_forward(feat)
            # no lags and local, get feature vector and loop over 2nd dimension
            else:
                X = np.array(X).reshape([n_points, len(X)])
                y = []
                for p in range(n_points):
                    y.append(feed_forward(X[p]))
                return np.array(y).flatten()

    def get_training_data(self, feats, target, lags=None, local=False, test_frac=0.0):
        """
        Generate trainig data. Training data can be made (time) lagged and/or local.

        Parameters
        ----------
        feats : Array or list of arrays
            A single feature array or a list of different feature arrays. The shape of the feature
            arrays must be (n_samples, n_points_i). Here n_samples is the number of samples and
            n_points_i is the size of a single sample of the i-th feature.
        target : Array
            The target that must be predicted by the surrogate. The shape of this array must be
            (n_samples, n_target), where n_target is the size of a single sample.
        lags : list of lists, optional
            In the case of a time-lagged surrogate, this list contains the time lags of
            each distinct feature. Each time lag is specified as a list. Example: if
            feats = [X1, X2], and lags = [[1], [1,2]], then X_1 will get lagged by one time step
            and X_2 will get lagged by two time steps. The default is None.
        local : boolean, optional
            If false, each feature sample of n_points_i will be used as input. If true, the
            surrogate will be applied locally, meaning that n_points_i separate scalar input
            faetures will be extracted from each feature sample. If true, the size of all
            features and target must be the same: n_points_i (for all i) = n_target = a fixed
            number. The default is False.
        test_frac : float, optional
            The final fraction of the training data that is withheld from training.
            The default is 0.0, and it must be in [0.0, 1.0].

        Returns
        -------
        None.

        """

        if not isinstance(feats, list):
            feats = [feats]

        # the number of distinct feature arrays
        self.n_feat_arrays = len(feats)

        # flag if the surrogate is to be applied locally or not
        self.local = local

        # initialize storage for online training
        self.online_feats = [[] for i in range(self.n_feat_arrays)]
        self.online_target = []

        # number of training samples
        self.n_samples = feats[0].shape[0]
        # number of points in the computational grid
        self.n_points = feats[0].shape[1]
        # compute the size of the training set based on value of test_frac
        self.n_train = np.int(self.n_samples * (1.0 - test_frac))
        print('Using first %d/%d samples for training' % (self.n_train, self.n_samples))

        X = {}
        y = {}
        # use the entire row as a feature
        if not local:
            # list of features
            X[0] = [X_i[0:self.n_train] for X_i in feats]
            # the data
            y[0] = target[0:self.n_train]
        # do not use entire row as feature, apply surrogate locally along second dimension
        else:
            # create a separate training set for every grid point
            for i in range(self.n_points):
                X[i] = [X_i[0:self.n_train, i] for X_i in feats]
                y[i] = target[0:self.n_train, i].reshape([-1, 1])

        X_train = []
        y_train = []
        # No time-lagged training data
        if lags is not None:
            self.max_lag = np.max(list(chain(*lags)))
            print('Creating time-lagged training data...')
            # lag every training set in X and y
            for i in range(len(X)):
                X_train_i, y_train_i = self.lag_training_data(X[i], y[i], lags=lags)
                X_train.append(X_train_i)
                y_train.append(y_train_i)
            X_train = np.concatenate(X_train)
            y_train = np.concatenate(y_train)
            print('done')
        else:
            self.max_lag = 0
            # no time lag, just add every entry in X and y to an array
            self.foo = X
            # loop over all spatial points in the case of a local surrogate. For non-local
            # surrogates, len(X) = 1
            for i in range(len(X)):
                # if only one unique feature vector is used
                if len(X[i]) == 1:
                    X_train.append(np.array(X[i]).reshape([self.n_train, -1]))
                # if there are multiple feature vectors, concatenate them
                else:
                    X_train.append(np.concatenate(X[i], axis=1))
                y_train.append(y[i])
            X_train = np.concatenate(X_train)
            y_train = np.concatenate(y_train)

        return X_train, y_train

    def get_online_training_data(self, **kwargs):
        """
        Return the training data for a single online-learning step.

        Returns
        -------
        X_train : array
            The feature array.
        y_train : array
            The target array.

        """

        feats = self.online_feats
        target = np.array(self.online_target)

        # if lagged feature vectors are used
        if self.lags is not None:

            feats = [np.array(feat) for feat in feats]
            if not self.local:
                # create (time-lagged) training data from X
                X_train, y_train = self.lag_training_data(feats, target, lags=self.lags,
                                                          init_feats=False)
            else:
                X_train = []
                y_train = []
                # create a separate training set for every grid point
                for i in range(self.n_points):
                    X_i = [X_i[:, i] for X_i in feats]
                    y_i = target[:, i].reshape([-1, 1])
                    # create time-lagged data per gridpoint
                    X_train_i, y_train_i = self.lag_training_data(X_i, y_i, lags=self.lags,
                                                                  init_feats=False)
                    X_train.append(X_train_i)
                    y_train.append(y_train_i)
                X_train = np.concatenate(X_train)
                y_train = np.concatenate(y_train)

        # do not time lag data
        else:
            # make a single array where each row contains a concateated vector of all feature
            # vectors
            X_train = np.concatenate(feats, axis=1)
            X_train = X_train.reshape([-1, kwargs['n_in']])
            y_train = target.reshape([-1, kwargs['n_out']])

        return X_train, y_train

    def generate_online_training_data(self, feats, LR_before, LR_after, HR_before, HR_after):
        """
        Compute the features and the target data for an online training step. Results are
        stored internally, and used within the 'train_online' subroutine.

        Source:
        Rasp, "Coupled online learning as a way to tackle instabilities and biases
        in neural network parameterizations: general algorithms and Lorenz 96
        case study", 2020.

        Parameters
        ----------
        feats : array or list of arrays
            The input features
        LR_before : array
            Low resolution state at previous time step.
        LR_after : array
            Low resolution state at current time step.
        HR_before : array
            High resolution state at previous time step.
        HR_after : array
            High resolution state at current time step.

        Returns
        -------
        None.

        """

        # multiple faetures arrays are stored in a list. For consistency put a single
        # array also in a list.
        if isinstance(feats, np.ndarray):
            feats = [feats]

        # store input features
        for i in range(self.n_feat_arrays):
            self.online_feats[i].append(feats[i])

        # difference of the low res model between time n and time n+1
        delta_LR = LR_after - LR_before
        # the difference between the low res and high res model at time n
        delta_nudge = LR_before - HR_before
        # difference of the high res model between time n and time n+1
        delta_HR = HR_after - HR_before

        # assume that over a small time interval [n, n+1] we can write:
        # delta_HR = would_be_delta_without_nuding + delta_due_to_nudging
        delta_no_nudge_HR = delta_HR - delta_nudge / self.tau_nudge * self.dt_LR

        # compute correction from: delta_LR + correction = delta_no_nudge_HR
        correction = delta_no_nudge_HR - delta_LR

        # make the correction the target for the neural network. Divide by timestep
        # since update is LR += correction * dt
        self.online_target.append(correction / self.dt_LR)

        # remove oldest item from the online features is the window lendth is exceeded
        if len(self.online_feats[0]) > self.window_length:
            for i in range(self.n_feat_arrays):
                self.online_feats[i].pop(0)
            self.online_target.pop(0)

    def set_online_training_parameters(self, tau_nudge, dt_LR, window_length):
        """
        Stores parameters required for online training.

        Parameters
        ----------
        tau_nudge : float
            Nudging time scale.
        dt_LR : float
            Time step low resolution model.
        window_length : int
            The length of the moving window in which online features are stored.

        Returns
        -------
        None.

        """
        self.tau_nudge = tau_nudge
        self.dt_LR = dt_LR
        self.window_length = window_length

    def lag_training_data(self, X, y, lags, init_feats=True):
        """
        Create time-lagged supervised training data X, y

        Parameters:
            X: features. Either an array of dimension (n_samples, n_features)
               or a list of arrays of dimension (n_samples, n_features)

            y: training target. Array of dimension (n_samples, n_outputs)

            lags: list of lists, containing the integer values of lags
                  Example: if X=[X_1, X_2] and lags = [[1], [1, 2]], the first
                  feature array X_1 is lagged by 1 (time) step and the second
                  by 1 and 2 (time) steps.

        Returns:
            X_train, y_trains (arrays), of lagged features and target data. Every
            row of X_train is one (time) lagged feature vector. Every row of y_train
            is a target vector at the next (time) step
        """

        # compute the max number of lags in lags
        lags_flattened = list(chain(*lags))
        max_lag = np.max(lags_flattened)

        # total number of data samples
        n_samples = y.shape[0]

        # if X is one array, add it to a list anyway
        if isinstance(X, np.ndarray):
            tmp = []
            tmp.append(X)
            X = tmp

        # compute target data at next (time) step
        if y.ndim == 2:
            y_train = y[max_lag:, :]
        elif y.ndim == 1:
            y_train = y[max_lag:]
        else:
            print("Error: y must be of dimension (n_samples, ) or (n_samples, n_outputs)")
            return

        # a lag list must be specified for every feature in X
        if len(lags) != len(X):
            print('Error: no specified lags for one of the featutes in X')
            return

        # compute the lagged features
        C = []
        idx = 0
        for X_i in X:

            for lag in np.sort(lags[idx])[::-1]:
                begin = max_lag - lag
                end = n_samples - lag

                if X_i.ndim == 2:
                    C.append(X_i[begin:end, :])
                elif X_i.ndim == 1:
                    C.append(X_i[begin:end])
                else:
                    print("Error: X must contains features of dimension (n_samples, ) \
                          or (n_samples, n_features)")
                    return
            idx += 1

        # C is a list of lagged features, turn into a single array X_train
        X_train = C[0]

        if X_train.ndim == 1:
            X_train = X_train.reshape([y_train.shape[0], 1])

        for X_i in C[1:]:

            if X_i.ndim == 1:
                X_i = X_i.reshape([y_train.shape[0], 1])

            X_train = np.append(X_train, X_i, axis=1)

        # initialize the storage of features
        if init_feats:
            self.empty_feature_history(lags)

        return X_train, y_train

    def bin_data(self, y, n_bins):
        """
        Bin the data y in to n_bins non-overlapping bins

        Parameters
        ----------
        y:  array
            size (number of samples, number of variables): Data
        n_bins: int
             Number of (equidistant) bins to be used.

        Returns
        -------
        None.

        """

        n_samples = y.shape[0]

        if y.ndim == 2:
            n_vars = y.shape[1]
        else:
            n_vars = 1
            y = y.reshape([n_samples, 1])

        self.binnumbers = np.zeros([n_samples, n_vars]).astype('int')
        self.y_binned = {}
        self.y_binned_mean = {}
        y_idx_binned = np.zeros([n_samples, n_bins * n_vars])
        self.bins = {}
        self.n_vars = n_vars

        for i in range(n_vars):

            self.y_binned[i] = {}
            self.y_binned_mean[i] = {}

            bins = np.linspace(np.min(y[:, i]), np.max(y[:, i]), n_bins + 1)
            self.bins[i] = bins

            _, _, self.binnumbers[:, i] = \
                stats.binned_statistic(y[:, i], np.zeros(n_samples), statistic='count', bins=bins)

            unique_binnumbers = np.unique(self.binnumbers[:, i])

            offset = i * n_bins

            for j in unique_binnumbers:
                idx = np.where(self.binnumbers[:, i] == j)
                self.y_binned[i][j - 1] = y[idx, i]
                self.y_binned_mean[i][j - 1] = np.mean(y[idx, i])
                y_idx_binned[idx, offset + j - 1] = 1.0

        return y_idx_binned

    def empty_feature_history(self, lags):
        """
        Initialize an empty feat_history dict. This dict keeps track of the features
        arrays that were used up until 'max_lag + 1' steps ago.

        Parameters:

            lags: list of lists, containing the integer values of lags
                  Example: if X=[X_1, X_2] and lags = [[1], [1, 2]], the first
                  feature array X_1 is lagged by 1 (time) step and the second
                  by 1 and 2 (time) steps.
        """
        self.lags = []

        for l in lags:
            self.lags.append(np.sort(l)[::-1])

        # self.max_lag = np.max(list(chain(*lags)))

        self.feat_history = {}

        # the number of feature arrays that make up the total input feature vector
        # self.n_feat_arrays = len(lags)

        for i in range(self.n_feat_arrays):
            self.feat_history[i] = []

    def initial_condition_feature_history(self, feats, start=0):
        """
        The features can be lagged in time. Therefore, the initial condition of the
        time-lagged feature vector must be set up. The training data is used
        for this.

        Parameters
        ----------
        + feats : a list of the variables used to construct the time-lagged
        features: [var_1, var_2, ...]. Each var_i is an array such that
        var_i[0] gives the value of var_i at t_0, var_i[1] at t_1 etc.

        + start : the starting index of the training features. Default is 0.

        Returns
        -------
        None.

        """
        for i in range(self.max_lag):
            if not self.local:
                feat = [X_i[start + i] for X_i in feats]
            else:
                feat = [X_i[start + i].reshape([1, -1]) for X_i in feats]

            self.append_feat(feat)

    def append_feat(self, X):
        """
        Append the feature vectors in X to feat_history dict

        Parameters:

            X: features. Either an array of dimension (n_samples, n_features)
               or a list of arrays of dimension (n_samples, n_features)

        """

        # if X is one array, add it to a list anyway
        if isinstance(X, np.ndarray):
            X = [X]

        for i in range(self.n_feat_arrays):

            assert isinstance(
                X[i], np.ndarray), 'ERROR: Only numpy arrays are allowed as input features.'

            self.feat_history[i].append(X[i])

            # if max number of features is reached, remove first item
            if len(self.feat_history[i]) > self.max_lag + 1:
                self.feat_history[i].pop(0)

    def get_feat_history(self, **kwargs):
        """
        Return the features from the feat_history dict based on the lags
        specified in self.lags

        Returns:
            X_i: array of lagged features of dimension (feat1.size + feat2.size + ...,)
        """
        X_i = []

        idx = 0
        for i in range(self.n_feat_arrays):
            for lag in self.lags[idx]:
                begin = self.max_lag - lag
                current_feat = self.feat_history[i][begin]
                if current_feat.ndim == 1:
                    X_i.append(current_feat)
                elif current_feat.shape[1] == 1:
                    X_i.append(current_feat.flatten())
                else:
                    X_i.append(np.array([current_feat[0][kwargs['index']]]))
            idx += 1

        return np.array(list(chain(*X_i)))

    # def standardize_data(self, standardize_X=True, standardize_y=True):
    #     """
    #     Standardize the training data
    #     """

    #     if standardize_X:
    #         X_mean = np.mean(self.X, axis=0)
    #         X_std = np.std(self.X, axis=0)
    #     else:
    #         X_mean = 0.0
    #         X_std = 1.0

    #     if standardize_y:
    #         y_mean = np.mean(self.y, axis=0)
    #         y_std = np.std(self.y, axis=0)
    #     else:
    #         y_mean = 0.0
    #         y_std = 1.0

    #     return (self.X - X_mean) / X_std, (self.y - y_mean) / y_std

    # def recursive_moments(self, X_np1, mu_n, sigma2_n, N):
    #     """
    #     Recursive formulas for the mean and variance. Computes the new moments
    #     when given a new sample and the old moments.

    #     Arguments
    #     ---------
    #     + X_np1: a new sample of random variable X
    #     + mu_n: the sample mean of X, not including X_np1
    #     + sigma2_n: the variance of X, not including X_np1
    #     + N: the total number of samples thus far

    #     Returns
    #     -------
    #     The mean and variance, updated based on the new sample

    #     """
    #     mu_np1 = mu_n + (X_np1 - mu_n) / (N + 1)
    #     sigma2_np1 = sigma2_n + mu_n**2 - mu_np1**2 + (X_np1**2 - sigma2_n - mu_n**2) / (N + 1)

    #     return mu_np1, sigma2_np1

    # def estimate_embedding_dimension(self, y, N):

    #     for n in range(3, N+1):

    #         lags_n = range(1, n)
    #         lags_np1 = range(1, n+1)
    #         y_n, _ = self.lag_training_data(y, np.zeros(y.size), [lags_n])
    #         y_np1, _ = self.lag_training_data(y, np.zeros(y.size), [lags_np1])

    #         L = y_np1.shape[0]
    #         dist_n = np.zeros(L)
    #         dist_np1 = np.zeros(L)

    #         for l in range(L):
    #             # d = np.linalg.norm(y_n - y_n[l], axis = 0)
    #             d = np.sum((y_n - y_n[l])**2, axis=1)
    #             a, d_min = np.partition(d, 1)[0:2]
    #             # d = np.linalg.norm(y_np1 - y_np1[l], axis = 0)
    #             d = np.sum((y_np1 - y_np1[l])**2, axis=1)
    #             _, d_min2 = np.partition(d, 1)[0:2]

    #             dist_n[l] = d_min
    #             dist_np1[l] = d_min2

    #         test = np.abs((dist_n - dist_np1)/dist_n)

    #         print(len(lags_n))
    #         print(np.where(test > 20)[0].size)



"""
Deep active subspace surrogate.
"""

import numpy as np

from .DAS_Layer import DAS_Layer
from .Layer import Layer
from .NN import ANN


class DAS_network(ANN):
    """
    Deep active subspace surrogate.

    Method:
        Tripathy, Rohit, and Ilias Bilionis. "Deep active subspaces: A scalable
        method for high-dimensional uncertainty propagation."
        ASME 2019 International Design Engineering Technical Conferences and
        Computers and Information in Engineering Conference.
        American Society of Mechanical Engineers Digital Collection, 2019.
    """

    def __init__(self, X, y, d, alpha=0.001, decay_rate=1.0, decay_step=10**5,
                 beta1=0.9, beta2=0.999, lamb=0.0, n_out=1,
                 param_specific_learn_rate=True, loss='squared', activation='tanh',
                 activation_out='linear', n_softmax=0, n_layers=2, n_neurons=16,
                 bias=True, batch_size=1, save=True,
                 name='DAS', on_gpu=False,
                 standardize_X=True, standardize_y=True, **kwargs):
        """
        Initialize the Deep Active Subspace surrogate object.

        Parameters
        ----------
        X : array
            The input features.
        y : array
            The target data.
        d : int
            The assumed dimension of the active subspace.
        alpha : float, optional
            The learning rate. The default is 0.001.
        decay_rate : float, optional
            Factor multiplying the decay rate every decay_step iterations.
            The default is 1.0.
        decay_step : int, optional
            The number of training iterations after which decay_rate is lowered.
            The default is 10**5.
        beta1 : float, optional
            Momentum parameter controlling the moving average of the loss gradient.
            Used for the parameter-specific learning rate. The default is 0.9.
        beta2 : float, optional
            Parameter controlling the moving average of the squared gradient.
            Used for the parameter-specific learning rate. The default is 0.999.
        lamb : float, optional
            L2 weight regularization parameter. The default is 0.0.
        n_out : int, optional
            The number of output neurons. The default is 1.
        param_specific_learn_rate : boolean, optional
            Use parameter-specific learing rate. The default is True.
        loss : string, optional
            The name of the loss function. The default is 'squared'.
        activation : string, optional
            The name of the activation function of the hidden layers.
            The default is 'tanh'.
        activation_out : string, optional
            The name of the activation function of the output layer.
            The default is 'linear'.
        n_softmax : int, optional
            The number of softmax layers attached to the output. The default is 0.
        n_layers : int, optional
            The number of layers, not counting the input layer. The default is 2.
        n_neurons : int, optional
            The number of neurons per hidden layer. The default is 16.
        bias : boolean, optional
            Use a bias neuron. The default is True.
        batch_size : int, optional
            The size of the mini batch. The default is 1.
        save : boolean, optional
            Save the neural network to a pickle file after training.
            The default is True.
        name : string, optional
            The name of the neural network. The default is 'DAS'.
        on_gpu : boolean, optional
            Train the neural network on a GPU using cupy. NOT IMPLEMENTED IN THIS VERSION.
            The default is False.
        standardize_X : boolean, optional
            Standardize the features. The default is True.
        standardize_y : boolean, optional
            Standardize the target data. The default is True.


        Returns
        -------
        None.

        """
        # the dimension of the active subspace
        self.d = d

        # set all the common parameters via the parent ANN class,
        # but overwrite the init_network subsroutine in this class
        super().__init__(X, y, alpha=alpha, decay_rate=decay_rate,
                         decay_step=decay_step, beta1=beta1, beta2=beta2,
                         lamb=lamb, n_out=n_out, loss=loss, activation=activation,
                         activation_out=activation_out, n_softmax=n_softmax,
                         n_layers=n_layers, n_neurons=n_neurons, bias=bias,
                         batch_size=batch_size,
                         param_specific_learn_rate=param_specific_learn_rate,
                         save=save, on_gpu=on_gpu, name=name,
                         standardize_X=standardize_X, standardize_y=standardize_y,
                         **kwargs)

    def init_network(self, **kwargs):
        """
        Set up the network structure by creating the Layer objects and
        connecting them together. The second layer is a DAS_Layer.
        This subroutine overwrites a subroutine from the parent ANN class.

        Returns
        -------
        None.

        """

        self.layers = []
        
        n_hidden = self.n_layers
        if 'cumsum' in kwargs and kwargs['cumsum'] == True:
            self.n_layers += 1

        # add the input layer
        self.layers.append(Layer(self.n_in, 0, self.n_layers, 'linear',
                                 self.loss, False, batch_size=self.batch_size,
                                 lamb=self.lamb, on_gpu=self.on_gpu))

        # add the deep active subspace layer
        self.layers.append(DAS_Layer(self.d, self.n_layers, True,
                                     batch_size=self.batch_size))

        # add the hidden layers
        for r in range(2, n_hidden):
            self.layers.append(Layer(self.n_neurons, r, self.n_layers, self.activation,
                                     self.loss, self.bias, batch_size=self.batch_size,
                                     lamb=self.lamb, on_gpu=self.on_gpu))

        # add the output layer
        self.layers.append(
            Layer(
                self.n_out,
                r+1,
                self.n_layers,
                self.activation_out,
                self.loss,
                batch_size=self.batch_size,
                lamb=self.lamb,
                n_softmax=self.n_softmax,
                on_gpu=self.on_gpu,
                **kwargs))

        super().connect_layers()
        super().print_network_info()

    def feed_forward(self, X_i, batch_size=1):
        """
        Run the deep active subspace network forward.

        Parameters
        ----------
        X_i : array
            The feauture array, needs to have shape [batch size, number of features].
        batch_size : int, optional
            The bath size. The default is 1.

        Returns
        -------
        array
            The prediction of the deep active subspace neural network.
        """

        # set the features at the output of in the input layer
        self.layers[0].h = X_i.T

        for i in range(1, self.n_layers + 1):
            # compute the output on the layer using matrix-maxtrix multiplication
            self.layers[i].compute_output(batch_size)

        return self.layers[-1].h

    def batch(self, X_i, y_i, alpha=0.001, beta1=0.9, beta2=0.999):
        """
        Update the weights of the neural network and the weights of
        the Gram-Schmidt vectors using a mini batch.

        Parameters
        ----------
        X_i : array
            The input features of the mini batch.
        y_i : array
            The target data of the mini batch.
        alpha : float, optional
            The learning rate. The default is 0.001.
        beta1 : float, optional
            Momentum parameter controlling the moving average of the loss gradient.
            Used for the parameter-specific learning rate. The default is 0.9.
        beta2 : float, optional
            Parameter controlling the moving average of the squared gradient.
            Used for the parameter-specific learning rate. The default is 0.999.

        Returns
        -------
        None.

        """

        self.feed_forward(X_i, self.batch_size)
        # if jacobian = True, the Q weights will be optimized using y_grad_Q,
        # instead of L_grad_Q. Is harder to train, maybe remove this option later.
        self.back_prop(y_i, jacobian=False)

        for r in range(1, self.n_layers + 1):

            layer_r = self.layers[r]

            # Deep active subspace layer
            if r == 1:
                # momentum
                layer_r.V = beta1 * layer_r.V + (1.0 - beta1) * layer_r.L_grad_Q
                # moving average of squared gradient magnitude
                layer_r.A = beta2 * layer_r.A + (1.0 - beta2) * layer_r.L_grad_Q**2
            # standard layer
            else:
                # momentum
                layer_r.V = beta1 * layer_r.V + (1.0 - beta1) * layer_r.L_grad_W
                # moving average of squared gradient magnitude
                layer_r.A = beta2 * layer_r.A + (1.0 - beta2) * layer_r.L_grad_W**2

            # select learning rate
            if not self.param_specific_learn_rate:
                # same alpha for all weights
                alpha_i = alpha
            # param specific learning rate
            else:
                # RMSProp
                alpha_i = alpha / (np.sqrt(layer_r.A + 1e-8))

            # gradient descent update step with L2 regularization
            if self.lamb > 0.0:
                layer_r.W = (1.0 - layer_r.Lamb * alpha_i) * layer_r.W - alpha_i * layer_r.V
            # without regularization
            else:
                # Deep active subspace layer
                if r == 1:
                    # update the Q weights
                    layer_r.Q = layer_r.Q - alpha_i * layer_r.V
                    # compute the weights W(Q) via Gram Schmidt
                    layer_r.compute_weights()
                # standard layer
                else:
                    layer_r.W = layer_r.W - alpha_i * layer_r.V



"""
CLASS TO PERFORM ANALYSIS ON RESULTS FROM A CONVERGENT CROSS MAPPING SURROGATE.
"""
from .base import BaseAnalysis
import matplotlib.pyplot as plt
import numpy as np
from scipy import stats
from scipy.spatial import ConvexHull


class CCM_analysis(BaseAnalysis):
    """
    CCM analysis class
    """

    def __init__(self, ccm_surrogate, **kwargs):
        print('Creating CCM_analysis object')
        self.ccm_surrogate = ccm_surrogate

    def auto_correlation_function(self, X, max_lag):
        """
        Compute the autocorrelation of X over max_lag time steps

        Parameters:
            - X (array, size (N,)): the samples from which to compute the ACF
            - max_lag (int): the max number of time steps, determines max
              lead time

        Returns:
            - R (array): array of ACF values
        """
        return super().auto_correlation_function(X, max_lag)

    def cross_correlation_function(self, X, Y, max_lag):
        """
        Compute the crosscorrelation between X and Y over max_lag time steps

        Parameters:
            - X, Y (array, size (N,)): the samples from which to compute the CCF
            - max_lag (int): the max number of time steps, determines max
              lead time

        Returns:
            - C (array): array of CCF values
        """
        return super().cross_correlation_function(X, Y, max_lag)

    def get_pdf(self, X, Npoints=100):
        """
        Computes a kernel density estimate of the samples in X

        Parameters:
            - X (array): the samples
            - Npoints (int, default = 100): the number of points in the domain of X

        Returns:
            - domain (array of size (Npoints,): the domain of X
            - kde (array of size (Npoints,): the kernel-density estimate
        """
        return super().get_pdf(X, Npoints=Npoints)

    def plot_2D_binning_object(self):
        """
        Visual representation of a 2D binning object. Also shows the mapping
        between empty to nearest non-empty bins.

        Returns
        -------
        None.

        """
        if self.ccm_surrogate.n_feats != 2:
            print('Only works for N_c = 2')
            return

        fig = plt.figure()
        ax = fig.add_subplot(111, xlabel=r'conditioning variable 1',
                             ylabel=r'conditioning variable 2')

        # plot bins and (c1, c2) which corresponding to a r sample point
        ax.plot(self.ccm_surrogate.feats[:, 0],
                self.ccm_surrogate.feats[:, 1], '+', color='lightgray', alpha=0.3)
        ax.vlines(self.ccm_surrogate.bins[0], np.min(
            self.ccm_surrogate.feats[:, 1]), np.max(self.ccm_surrogate.feats[:, 1]))
        ax.hlines(self.ccm_surrogate.bins[1], np.min(
            self.ccm_surrogate.feats[:, 0]), np.max(self.ccm_surrogate.feats[:, 0]))

        ax.plot(self.ccm_surrogate.x_mid_pad_tensor[:, 0],
                self.ccm_surrogate.x_mid_pad_tensor[:, 1], 'g+')

        # plot the mapping
        for i in range(self.ccm_surrogate.max_binnumber):
            ax.plot([self.ccm_surrogate.x_mid_pad_tensor[i][0],
                     self.ccm_surrogate.x_mid_pad_tensor[self.ccm_surrogate.mapping[i]][0]],
                    [self.ccm_surrogate.x_mid_pad_tensor[i][1],
                     self.ccm_surrogate.x_mid_pad_tensor[self.ccm_surrogate.mapping[i]][1]],
                    'b',
                    alpha=0.6)

        ax.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))
        ax.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
        plt.tight_layout()
        plt.show()

    def plot_2D_shadow_manifold(self):
        """
        Plots the bins on the manifold of the conditioning variables, and
        the corresponding neighborhoods on the shadow manifold. Click on
        manifold bin to show the corresponding shadow bin.

        Returns
        -------
        None.

        """

        if self.ccm_surrogate.n_feats == 1:
            print('Dimension manifold = 1, only works for dimension of 2 or higher')
            return
        elif self.ccm_surrogate.n_feats > 2:
            print('Dimension manifold > 2, will only plot first 2 dimensions')

        print('Click on the bins on the left to show the corresponding shadow bins.')

        fig = plt.figure('manifolds_and_binnumbers', figsize=[8, 4])
        # colors = cycle(['--r', '--g', '--b', '--m', '--k'])
        self.ax1 = fig.add_subplot(121, title='Manifold')
        self.ax2 = fig.add_subplot(122, title='Shawdow manifold')
        self.ax1.set_xlabel(r'$\mathrm{conditioning\;variable\;at\;t-\tau}$')
        self.ax1.set_ylabel(r'$\mathrm{conditioning\;variable\;at\;t}$')
        self.ax2.set_xlabel(r'$\mathrm{shadow\;variable\;at\;t-\tau}$')
        self.ax2.set_ylabel(r'$\mathrm{shadow\;variable\;at\;t}$')

        # plot the samples
        self.ax1.plot(self.ccm_surrogate.feats[:, 0], self.ccm_surrogate.feats[:, 1],
                      '+', color='lightgray', alpha=0.3)
        self.ax2.plot(self.ccm_surrogate.target[:, 0], self.ccm_surrogate.target[:, 1],
                      '+', color='lightgray', alpha=0.3)

        for idx_i in self.ccm_surrogate.sample_idx_per_bin.keys():

            idx = self.ccm_surrogate.sample_idx_per_bin[idx_i]
            points_c = self.ccm_surrogate.feats[idx, 0:2]
            points_r = self.ccm_surrogate.target[idx, 0:2]

            self.plot_local_bin(points_c, idx_i, self.ax1, '--k')
            self.plot_local_bin(points_r, idx_i, self.ax2, '--k')

        plt.tight_layout()
        cid = fig.canvas.mpl_connect('button_press_event', self.onclick)

    def onclick(self, event):
        """
        Evenht handler for mouse click in the manifold plot

        Parameters
        ----------
        event : button_press_event.

        Returns
        -------
        None.

        """

        if event.inaxes != self.ax1:
            return

        # get the location of the click
        c_i = np.array([event.xdata, event.ydata]).reshape([1, 2])
        # find corresponding bin index
        _, _, binnumber_i = stats.binned_statistic_dd(c_i, np.zeros(1),
                                                      bins=self.ccm_surrogate.bins)

        print('Bin', binnumber_i[0])
        # map empty bins to nearest non-empty bin
        binnumber_i = self.ccm_surrogate.mapping[binnumber_i][0]
        # find the indices of the samples of the selected bin
        idx = self.ccm_surrogate.sample_idx_per_bin[binnumber_i]
        # find the corresponding points of the features and the data
        points_c = self.ccm_surrogate.feats[idx]
        points_r = self.ccm_surrogate.target[idx]

        # highlight the selected bin + the corresponding shadow bin
        self.plot_local_bin(points_c, binnumber_i, self.ax1, 'r', width=4)
        self.plot_local_bin(points_r, binnumber_i, self.ax2, 'r', width=4)

        plt.draw()

    def plot_local_bin(self, points, binnumber, ax, marker, width=2):
        """
        Highlight a bin on the shadow manifold plot

        Parameters
        ----------
        points : points in a given bin
        binnumber : the corresponding binnumber
        ax : matplotlib axis
        marker : matplotlib marker
        width : line width

        Returns
        -------
        None.

        """
        # if  there are 3 or more samples in current bin, plot the
        # convex hull of all samples in this bin
        if points.shape[0] >= 3:
            # marker = next(colors)
            hull = ConvexHull(points)
            ax.plot(points[hull.vertices, 0], points[hull.vertices, 1], marker, linewidth=width)
            ax.plot([points[hull.vertices[0], 0], points[hull.vertices[-1], 0]],
                    [points[hull.vertices[0], 1], points[hull.vertices[-1], 1]], marker, linewidth=width)
        # print(hull.volume)
        #     #plot the binnumber
        #     x_mid = np.mean(points[hull.vertices, :], axis=0)
        #     ax.text(x_mid[0], x_mid[1], str(binnumber))
        # # ax.plot(X[idx_i, 0], X[idx_i, 1], '+')
        # else:
        #     x_mid = np.mean(points, axis=0)
        #     ax.text(x_mid[0], x_mid[1], str(binnumber))

    def compare_convex_hull_volumes(self):
        """
        Display some statistics on the convex hull volumes of the bins
        of the manifold and the shadow manifold.
        """

        self.vols = []
        self.ratio_vols = []

        # total volume of the featue point cloud and the data cloud
        total_vol_c = ConvexHull(self.ccm_surrogate.feats).volume
        total_vol_r = ConvexHull(self.ccm_surrogate.target).volume

        weights = []

        # loop over all non-empty bin indices
        for idx_i in self.ccm_surrogate.sample_idx_per_bin.keys():
            # get feature points ana corresp data points
            idx = self.ccm_surrogate.sample_idx_per_bin[idx_i]
            points_c = self.ccm_surrogate.feats[idx]
            points_r = self.ccm_surrogate.target[idx]

            # if there are enough points to make a ND convex hull
            if points_c.shape[0] >= self.ccm_surrogate.n_feats + 1:
                # Convex hull of local features / data points
                hull_c = ConvexHull(points_c)
                hull_r = ConvexHull(points_r)
                # volumes fractions
                vol_fraction_c = hull_c.volume / total_vol_c
                vol_fraction_r = hull_r.volume / total_vol_r

                weights.append(vol_fraction_c)
                # volume frac and ration of vol fracs
                self.vols.append([vol_fraction_c, vol_fraction_r])
                self.ratio_vols.append(vol_fraction_c / vol_fraction_r)

        avg_ratio = np.mean(self.ratio_vols)

        # weights = weights/np.sum(weights)
        # print(weights)
        # print(np.sum(weights))
        # weighted_ratio = np.mean(weights*self.ratio_vols)
        # self.weights = weights

        print('Average volume ratio binning cell/shadow cell = %.3f' % avg_ratio)
        print('Max volume ratio binning cell/shadow cell = %.3f' % np.max(self.ratio_vols))
        print('min volume ratio binning cell/shadow cell = %.3f' % np.min(self.ratio_vols))
        # print('Weighted average volume ratio binning cell/shadow cell = %.3f' % weighted_ratio)



"""
The Base analysis class which all other analysis classes inherit. Contains
general functions which can be of use for various analysis objects, e.g.
computing a kernel density estimate.
"""

import numpy as np
from sklearn.neighbors import KernelDensity


class BaseAnalysis:

    def __init__(self):
        """
        Constructor
        """
        self.name = 'Base_Analysis'

    def auto_correlation_function(self, X, max_lag):
        """
        Compute the autocorrelation of X over max_lag time steps

        Parameters:
            - X (array, size (N,)): the samples from which to compute the ACF
            - max_lag (int): the max number of time steps, determines max
              lead time

        Returns:
            - R (array): array of ACF values
        """
        lags = np.arange(1, max_lag)
        R = np.zeros(lags.size)

        idx = 0

        print('Computing auto-correlation function')

        # for every lag, compute autocorrelation:
        # R = E[(X_t - mu_t)*(X_s - mu_s)]/(std_t*std_s)
        for lag in lags:

            X_t = X[0:-lag]
            X_s = X[lag:]

            mu_t = np.mean(X_t)
            std_t = np.std(X_t)
            mu_s = np.mean(X_s)
            std_s = np.std(X_s)

            R[idx] = np.mean((X_t - mu_t) * (X_s - mu_s)) / (std_t * std_s)
            idx += 1

        print('done')

        e_fold_idx = np.where(R <= np.e**-1)[0][0]
        print('E-folding index = %d' % e_fold_idx)

        return R

    def cross_correlation_function(self, X, Y, max_lag):
        """
        Compute the crosscorrelation between X and Y over max_lag time steps

        Parameters:
            - X, Y (array, size (N,)): the samples from which to compute the CCF
            - max_lag (int): the max number of time steps, determines max
              lead time

        Returns:
            - C (array): array of CCF values
        """
        lags = np.arange(1, max_lag)
        C = np.zeros(lags.size)

        idx = 0

        print('Computing cross-correlation function')

        # for every lag, compute cross correlation:
        # R = E[(X_t - mu_Xt)*(Y_s - mu_Ys)]/(std_Xt*std_Ys)
        for lag in lags:

            X_t = X[0:-lag]
            Y_s = Y[lag:]

            mu_t = np.mean(X_t)
            std_t = np.std(X_t)
            mu_s = np.mean(Y_s)
            std_s = np.std(Y_s)

            C[idx] = np.mean((X_t - mu_t) * (Y_s - mu_s)) / (std_t * std_s)
            idx += 1

        print('done')

        return C

    def get_pdf(self, X, Npoints=100):
        """
        Computes a kernel density estimate of the samples in X

        Parameters:
            - X (array): the samples
            - Npoints (int, default = 100): the number of points in the domain of X

        Returns:
            - domain (array of size (Npoints,): the domain of X
            - kde (array of size (Npoints,): the kernel-density estimate
        """

        #    kernel = stats.gaussian_kde(X, bw_method='scott')
        #    x = np.linspace(np.min(X), np.max(X), Npoints)
        #    pde = kernel.evaluate(x)
        #    return x, pde

        print('Computing kernel-density estimate')

        X_min = np.min(X)
        X_max = np.max(X)
        bandwidth = (X_max - X_min) / 40

        kde = KernelDensity(kernel='gaussian', bandwidth=bandwidth).fit(X.reshape(-1, 1))
        domain = np.linspace(X_min, X_max, Npoints).reshape(-1, 1)
        log_dens = kde.score_samples(domain)

        print('done')

        return domain, np.exp(log_dens)

    def recursive_moments(self, X_np1, mu_n, sigma2_n, N):
        """
        Update the mean and variance using a new data sample X_np1.

        Parameters
        ----------
        X_np1 : aray
            A new data point, iteration n+1
        mu_n : array
            The mean compute using all samples up to iteration n
        sigma2_n : array
            The variance compute using all samples up to iteration n.
        N : int
            The number of samples.

        Returns
        -------
        mu_np1 : array
            The new mean, updated with the latest data point.
        sigma2_np1 : TYPE
            The new variance, updated with the latest data point..

        """
    
        mu_np1 = mu_n + (X_np1 - mu_n)/(N+1)
    
        sigma2_np1 = sigma2_n + mu_n**2 - mu_np1**2 + (X_np1**2 - sigma2_n - mu_n**2)/(N+1)
    
        return mu_np1, sigma2_np1


"""
CLASS FOR A CONVERGENT CROSS MAPPING SURROGATE
"""

import numpy as np
from scipy import stats
from itertools import chain, product, cycle
import matplotlib.pyplot as plt
import sys
from scipy.spatial import ConvexHull
import easysurrogate as es


class CCM_Surrogate:

    def __init__(self, **kwargs):
        print('Creating Convergent Cross-Mapping Object')
        self.name = 'CCM Surrogate'

    ############################
    # START COMMON SUBROUTINES #
    ############################

    def train(self, feats, target, N_bins, lags, min_count=0, test_frac=0):

        # spatial dimension, hardcoded to 1 for now
        self.N = 1

        # Feature engineering object
        self.feat_eng = es.methods.Feature_Engineering()

        # number of training samples
        self.n_samples = feats[0].shape[0]
        # compute the size of the training set based on value of test_frac
        self.n_train = np.int(self.n_samples * (1.0 - test_frac))
        print('Using first', self.n_train, 'of', self.n_samples, 'samples to train QSN')

        # list of features
        X = [X_i[0:self.n_train] for X_i in feats]
        # the data
        y = [target[0:self.n_train]]

        print('Creating time-lagged training data...')
        X_lagged, _ = self.feat_eng.lag_training_data(X, np.zeros(self.n_train), lags)
        Y_lagged, _ = self.feat_eng.lag_training_data(y, np.zeros(self.n_train), lags)
        print('done')

        # total number of conditional variables including lagged terms
        self.n_feats = X_lagged.shape[1]

        # number of unique conditioning variables (not including lags)
        self.N_covar = len(lags)

        self.feats = X_lagged
        self.target = Y_lagged
        self.N_bins = N_bins
        self.lags = lags
        self.max_lag = np.max(list(chain(*lags)))
        self.covar = {}

        for i in range(self.N_covar):
            self.covar[i] = []

        bins = self.get_bins(N_bins)

        count, binedges, binnumber = stats.binned_statistic_dd(self.feats, np.zeros(self.n_samples),
                                                               statistic='count', bins=bins)

        # the unique set of binnumers which have at least one r_ip1 sample
        unique_binnumbers = np.unique(binnumber)

        # some scalars
        binnumber_max = np.max(unique_binnumbers)

        # array containing r_ip1 indices sorted PER BIN
        # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # SHOULD BE A 1D ARRAY, AN ARRAY OF SIZE [MAX(BINNUMBER), MAX(COUNT)]
        # WILL STORE MOSTLY ZEROS IN HIGHER DIMENSIONS, LEADING TO MEMORY FAILURES
        # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        idx_of_bin = []

        # number of samples in each bin
        size_of_bin = []

        for b in unique_binnumbers:
            # r_ip1 indices of bin b
            tmp = np.where(binnumber == b)[0]

            idx_of_bin.append(tmp)
            size_of_bin.append(tmp.size)

        # make a 1d array of the nested arrays
        idx_of_bin = np.array(list(chain(*idx_of_bin)))
        size_of_bin = np.array(size_of_bin)

        # the starting offset of each bin in the 1D array idx_of_bin
        # offset[unique_binnumbers] will give the starting index in idex_of_bin for
        # each bin. Some entries are zero, which correspond to empty bins (except for
        # the lowest non-empty binnumber)
        offset = np.zeros(unique_binnumbers.size).astype('int')
        offset[1:] = np.cumsum(size_of_bin[0:-1])
        tmp = np.zeros(binnumber_max + 1).astype('int')
        tmp[unique_binnumbers] = offset
        self.offset = tmp

        ####################
        #empty bin handling#
        ####################

        # find indices of empty bins
        # Note: if 'full' is defined as 1 or more sample, binnumbers_nonempty is
        # the same as binnumbers_unique
        x_idx_nonempty = np.where(count > min_count)
        x_idx_nonempty_p1 = [x_idx_nonempty[i] + 1 for i in range(self.n_feats)]
        binnumbers_nonempty = np.ravel_multi_index(x_idx_nonempty_p1, [len(b) + 1 for b in bins])
        N_nonempty = binnumbers_nonempty.size

        # mid points of the bins
        x_mid = [0.5 * (bins[i][1:] + bins[i][0:-1]) for i in range(self.n_feats)]

        # mid points of the non-empty bins
        midpoints = np.zeros([N_nonempty, self.n_feats])
        for i in range(self.n_feats):
            midpoints[:, i] = x_mid[i][x_idx_nonempty[i]]

        #########################################
        # find all sample indices per bin index #
        #########################################

        sample_idx_per_bin = {}
        for j in unique_binnumbers:
            sample_idx_per_bin[j] = np.where(binnumber == j)[0]

        self.bins = bins
        self.count = count
        self.binnumber = binnumber
        self.binnumbers_nonempty = binnumbers_nonempty
        self.midpoints = midpoints
        self.idx_of_bin = idx_of_bin
        self.unique_binnumbers = unique_binnumbers
        self.sample_idx_per_bin = sample_idx_per_bin

        print('Connecting all possible empty bins to nearest non-empty bin...')
        self.fill_in_blanks()
        print('done.')

        self.init_feature_history(feats)

    def predict(self, X, stochastic=False):

        if not isinstance(X, list):
            X = [X]
        # append the current state X to the feature history
        self.feat_eng.append_feat(X)
        # get the feature history defined by the specified number of time lags.
        # Here, feat is an array with the same size as the neural network input layer
        feat = self.feat_eng.get_feat_history()
        feat = feat.reshape([1, self.n_feats])

        # find in which bins the c_i samples fall
        _, _, binnumbers_i = stats.binned_statistic_dd(feat, np.zeros(self.N), bins=self.bins)

        # static correction for outliers, using precomputed mapping array
        binnumbers_i = self.mapping[binnumbers_i]

        # the neighbors in the selected bin, on the 'C manifold'
        neighbors = self.feats[self.sample_idx_per_bin[binnumbers_i[0]]]

        # the distance from the current point to all neighbors
        dists = np.linalg.norm(neighbors - feat, axis=1)

        # if the current bin does not contain enough points to form a simplex,
        # adjust K.
        if dists.size < self.n_feats + 1:
            K = dists.size  # adjusted K
        else:
            K = self.n_feats + 1  # simplex K (manifold dimension + 1)

        # sort the distances + select the K nearest neighbors
        idx = np.argsort(dists)[0:K]
        simplex_idx = self.sample_idx_per_bin[binnumbers_i[0]][idx]

        if not stochastic:
            # compute the simplex weights w_i
            d_i = dists[idx]
            if d_i[0] == 0:
                u_i = np.zeros(K)
                u_i[0] = 1.0
            else:
                u_i = np.exp(-d_i / d_i[0])
            w_i = u_i / np.sum(u_i)

            # prediction is the weighted sum of correspoding samples
            # on the showdow manifold
            shadow_sample = np.sum(self.target[simplex_idx, -1] * w_i)

            return shadow_sample
        else:
            shadow_sample = self.sample_simplex(self.target[simplex_idx])[0]
            return shadow_sample[-1]

    def save_state(self):
        """
        Save the state of the QSN surrogate to a pickle file
        """
        state = self.__dict__
        super().save_state(state=state, name=self.name)

    def load_state(self):
        """
        Load the state of the QSN surrogate from file
        """
        super().load_state(name=self.name)

    ##########################
    # END COMMON SUBROUTINES #
    ##########################

    def init_feature_history(self, feats, start=0):
        """
        The features are assumed to be lagged in time. Therefore, the initial
        time-lagged feature vector must be set up. The training data is used
        for this.

        Parameters
        ----------
        + feats : a list of the variables used to construct the time-lagged
        features: [var_1, var_2, ...]. Each var_i is an array such that
        var_i[0] gives the value of var_i at t_0, var_i[1] at t_1 etc.

        + start : the starting index of the training features. Default is 0.

        Returns
        -------
        None.

        """
        for i in range(self.max_lag):
            feat = [X_i[start + i] for X_i in feats]
            self.feat_eng.append_feat(feat)

    def sample_simplex(self, xi, n_mc=1):
        """
        Use an analytical function map to points in the n_xi-dim. hypercube to a
        n_xi-dim simplex with vertices xi_k_jl.

        Source:
            W. Edeling, R. Dwight, P. Cinnella,
            Simplex-stochastic collocation method with improved scalability
            Journal of Computational Physics, 310, 301--328, 2016.

        Parameters
        ----------
        xi: array of floats, shape (n_xi + 1, n_xi)

        n_mc : int, number of samples to draw from inside the simplex
            DESCRIPTION. The default is 1.

        Returns
        -------
        n_mc uniformly distributed points inside the simplex.

        """
        n_xi = xi.shape[1]
        P = np.zeros([n_mc, n_xi])
        for k in range(n_mc):
            # random points inside the hypercube
            r = np.random.rand(n_xi)

            # the term of the map is \xi_k_j0
            sample = np.copy(xi[0])
            for i in range(1, n_xi + 1):
                prod_r = 1.
                # compute the product of r-terms: prod(r_{n_xi-j+1}^{1/(n_xi-j+1)})
                for j in range(1, i + 1):
                    prod_r *= r[n_xi - j]**(1. / (n_xi - j + 1))
                # compute the ith term of the sum: prod_r*(\xi_i-\xi_{i-1})
                sample += prod_r * (xi[i] - xi[i - 1])
            P[k, :] = sample

        return P

    # check which c_i fall within empty bins and correct binnumbers_i by
    # projecting to the nearest non-empty bin
    def check_outliers(self, binnumbers_i, c_i):

        # find out how many BINS with outliers there are
        unique_binnumbers_i = np.unique(binnumbers_i)
        idx = np.where(np.in1d(unique_binnumbers_i, self.binnumbers_nonempty) == False)[0]
        N_outlier_bins = idx.size

        if N_outlier_bins > 0:

            # index of outlier SAMPLES in binnumbers_i
            outliers_idx = np.in1d(binnumbers_i, unique_binnumbers_i[idx]).nonzero()[0]
            N_outliers = outliers_idx.size

#            if self.verbose == True:
#                print(N_outlier_bins, ' bins with', N_outliers ,'outlier samples found')

            # x location of outliers
            x_outliers = np.copy(c_i[outliers_idx])

            # find non-empty bin closest to the outliers
            closest_idx = np.zeros(N_outliers).astype('int')
            for i in range(N_outliers):
                if self.n_feats == 1:
                    dist = np.linalg.norm(self.midpoints - x_outliers[i], 2, 1)
                else:
                    dist = np.linalg.norm(self.midpoints - x_outliers[i, :], 2, 1)
                closest_idx[i] = np.argmin(dist)

            binnumbers_closest = self.binnumbers_nonempty[closest_idx]

            check = self.binnumbers_nonempty[closest_idx]
            x_idx_check = np.unravel_index(check, [len(b) + 1 for b in self.bins])
            x_idx_check = [x_idx_check[i] - 1 for i in range(self.n_feats)]

            # overwrite outliers in binnumbers_i with nearest non-empty binnumber
            binnumbers_closest = self.binnumbers_nonempty[closest_idx]

#            if self.verbose == True:
#                print('Moving', binnumbers_i[outliers_idx], '-->', binnumbers_closest)

            binnumbers_i[outliers_idx] = binnumbers_closest

    def fill_in_blanks(self):
        """
        Create an a-priori mapping between every possible bin and the nearest
        non-empty bin. Non-empty bins will link to themselves.

        Returns
        -------
        None.

        """

        bins_padded = []

        # mid points of all possible bins, including ghost bins
        for i in range(self.n_feats):
            dx1 = self.bins[i][1] - self.bins[i][0]
            dx2 = self.bins[i][-1] - self.bins[i][-2]

            # pad the beginning and end of current 1D bin with extrapolated values
            bin_pad = np.pad(self.bins[i], (1, 1), 'constant', constant_values=(
                self.bins[i][0] - dx1, self.bins[i][-1] + dx2))
            bins_padded.append(bin_pad)

        # compute the midpoints of the padded bins
        x_mid_pad = [0.5 * (bins_padded[i][1:] + bins_padded[i][0:-1]) for i in range(self.n_feats)]
        self.x_mid_pad_tensor = np.array(list(product(*x_mid_pad)))

        # total number bins
        self.max_binnumber = self.x_mid_pad_tensor.shape[0]

        mapping = np.zeros(self.max_binnumber).astype('int')

        for i in range(self.max_binnumber):

            if np.mod(i, 10) == 0:
                progress(i, self.max_binnumber)

            # bin is nonempty, just use current idx
            # if np.in1d(i, self.unique_binnumbers) == True:
            if np.in1d(i, self.binnumbers_nonempty):
                mapping[i] = i
            # bin is empty, find nearest non-empty bin
            else:
                binnumbers_i = np.array([i])
                self.check_outliers(
                    binnumbers_i, self.x_mid_pad_tensor[i].reshape([1, self.n_feats]))
                mapping[i] = binnumbers_i[0]

        self.mapping = mapping

    def print_bin_info(self):
        print('-------------------------------')
        print('Total number of samples= ', self.r_ip1.size)
        print('Total number of non-empty bins = ', self.binnumbers_nonempty.size)
        print('-------------------------------')

    # compute the uniform bins of the conditional variables in c
    def get_bins(self, N_bins):

        bins = []

        for i in range(self.n_feats):
            bins.append(np.linspace(np.min(self.feats[:, i]),
                                    np.max(self.feats[:, i]), N_bins[i] + 1))

        return bins


def progress(count, total, status=''):
    """
    progress bar for the command line
    Source: https://gist.github.com/vladignatyev/06860ec2040cb497f0f3
    """
    bar_len = 60
    filled_len = int(round(bar_len * count / float(total)))

    percents = round(100.0 * count / float(total), 1)
    bar = '=' * filled_len + '-' * (bar_len - filled_len)

    sys.stdout.write('[%s] %s%s ...%s\r' % (bar, percents, '%', status))
    sys.stdout.flush()  # As suggested by Rom Ruben (see: http://stackoverflow.com/questions/3173320/text-progress-bar-in-the-console/27871113#comment50529068_27871113)



"""
CLASS TO PERFORM ANALYSIS ON RESULTS FROM A KERNEL MIXTURE SURROGATE.
"""
from .base import BaseAnalysis
import os
import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import norm
from matplotlib import animation
# from sklearn.neighbors import KernelDensity


class KMN_analysis(BaseAnalysis):
    """
    KMN analysis class
    """

    def __init__(self, kmn_surrogate, **kwargs):
        print('Creating KMN_analysis object')
        self.kmn_surrogate = kmn_surrogate

    def compute_kde(self, dom, w, mu, sigma):
        """
        Compute a kernel density estimate (KDE) given the weights w, which are the output
        of the KMN surrogate.

        Parameters:
            - dom (array): domain of the KDE
            - w (array): weights of the KDE, as predicted by the network. Sum to 1.
        """

        K = norm.pdf(dom, mu, sigma)
        w = w.reshape([w.size, 1])

        return np.sum(w * K, axis=0)

    def make_movie(self, n_frames=500):
        """
        Makes a move using the training data. Left subplot shows the evolution of the
        kernel-density estimate, and the right subplot show the time series of the data
        and the random samples drawm from the kernel-density estimate. Saves the movie
        to a .gif file.

        Parameters

        n_frames (int): default is 500
            The number of frames to use in the movie.

        Returns: None
        """

        # get the (normalized, time-lagged) training data from the neural network
        X = self.kmn_surrogate.neural_net.X
        y = self.kmn_surrogate.neural_net.y

        print('===============================')
        print('Making movie...')

        # list to store the movie frames in
        ims = []
        fig = plt.figure(figsize=[8, 4])
        ax1 = fig.add_subplot(121, xlabel=r'$B_k$', ylabel=r'', yticks=[])
        ax2 = fig.add_subplot(122, xlabel=r'time', ylabel=r'$B_k$')
        plt.tight_layout()

        # number of features
        n_feat = X.shape[1]
        # number of softmax layers
        n_softmax = self.kmn_surrogate.n_softmax
        # kernel properties, mean and standard dev
        mu = self.kmn_surrogate.kernel_means
        sigma = self.kmn_surrogate.kernel_stds
        # number of points to use for plotting the conditional pdfs
        n_kde = 100

        # domain of the conditional pdfs
        dom = np.linspace(np.min(y), np.max(y), n_kde)

        # allocate memory
        kde = np.zeros([n_kde, n_softmax])
        samples = np.zeros([n_frames, n_softmax])

        # make movie by evaluating the network at TRAINING inputs
        for i in range(n_frames):

            # draw a random sample from the network
            w, idx_max, _ = self.kmn_surrogate.neural_net.get_softmax(X[i].reshape([1, n_feat]))
            for j in range(n_softmax):
                kde[:, j] = self.compute_kde(dom, w[j], mu[j], sigma[j])
                samples[i, j] = norm.rvs(mu[j][idx_max[j]], sigma[j][idx_max[j]])
            if np.mod(i, 100) == 0:
                print('i =', i, 'of', n_frames)

            # create a single frame, store in 'ims'
            plt2 = ax1.plot(dom, kde[:, 0] / np.max(kde[:, 0]), 'b', label=r'conditional pdf')
            plt3 = ax1.plot(y[i][0], 0.0, 'ro', label=r'data')
            plt4 = ax2.plot(y[0:i, 0], 'ro', label=r'data')
            plt5 = ax2.plot(samples[0:i, 0], 'g', label='random sample')

            if i == 0:
                ax1.legend(loc=1, fontsize=9)
                ax2.legend(loc=1, fontsize=9)

            ims.append((plt2[0], plt3[0], plt4[0], plt5[0],))

        # make a movie of all frame in 'ims'
        im_ani = animation.ArtistAnimation(fig, ims, interval=20,
                                           repeat_delay=2000, blit=True)
        im_ani.save('kvm.gif')
        print('done. Saved movie to kvm.gif')



import numpy as np
from scipy import stats
from itertools import chain


class SimpleBin:

    def __init__(self, feat_eng):
        """
        Parameters
        ----------
        feat_eng : EasySurrogate Feature_Engineering object

        Returns
        -------
        None.

        """

        if not hasattr(feat_eng, 'y_binned'):
            print("Error: feat_eng object does not contain binned data")
            print("Run the bin_data(..) subroutine of feat_eng")
            return

        # number of variables
        self.n_vars = feat_eng.n_vars

        # dict of the binned data. To access the binned data of bin 1 of the
        # 6-th variable use: y_binned[5][1]
        self.y_binned = feat_eng.y_binned
        self.y_binned_mean = feat_eng.y_binned_mean

    def resample(self, bin_idx):
        """
        Resamples reference data from bins specified by bin indices bin_idx.
        Bin indices are integers >= 1.

        Parameters
        ----------
        bin_idx : array of integers, size (nvars,): the bin indices of each
                  output variable

        Returns
        -------
        pred : array of floats, size (nvars,): array of resampled reference
               data. Samples are drawn from the bins specified by bin_idx

        """

        pred = np.zeros(self.n_vars)

        for i in range(self.n_vars):
            pred[i] = np.random.choice(self.y_binned[i][bin_idx[i]][0])

        return pred

    def resample_mean(self, bin_idx):
        """
        Resamples reference bin mean specified by bin indices bin_idx.
        Bin indices are integers >= 1.

        Parameters
        ----------
        bin_idx : array of integers, size (nvars,): the bin indices of each
                  output variable

        Returns
        -------
        pred : array of floats, size (nvars,): array of resampled bin mean
               data.

        """

        pred = np.zeros(self.n_vars)

        for i in range(self.n_vars):
            pred[i] = self.y_binned_mean[i][bin_idx[i]]

        return pred



#from .resampling import Resampler
from .NN import ANN
from .SimpleBin import SimpleBin
from .Feature_Engineering import Feature_Engineering
#from .RNN import RNN
#from .Input_Layer import Input_Layer
from .ccm_surrogate import CCM_Surrogate
from .qsn_surrogate import QSN_Surrogate
from .kmn_surrogate import KMN_Surrogate
from .reduced_surrogate import Reduced_Surrogate
from .ann_surrogate import ANN_Surrogate
from .DAS_network import DAS_network
from .CumSum_Layer import CumSum_Layer


from os import path
from setuptools import setup, find_packages

setup(
    name='easysurrogate',

    version=0.13,

    description=(''),

#    url='https://readthedocs.org/projects/easyvvuq/',

    author='CWI',

    install_requires=['numpy', 'scipy', 'matplotlib', 'h5py', 'sklearn'],

    packages=find_packages(),

    include_package_data=True,
)



"""
EasySurrogate campaign class
"""

from tkinter import filedialog
import tkinter as tk
import pickle
import h5py


class Campaign:
    """
    The main EasySurrogate Campaign object
    """

    def __init__(self, name=None, load_state=False, **kwargs):
        """
        Create a Campaign object.

        Parameters
        ----------
        name : string, optional
            The name of this object. The default is None.
        load_state : boolean, optional
            Load a campaign state from file. The default is False.

        Returns
        -------
        None.

        """

        if load_state:
            self.load_state(**kwargs)

        if name is None:
            self.campaign_name = 'EasySurrogate_Campaign'
        else:
            self.campaign_name = name

        self.accum_data = {}

    def load_hdf5_data(self, **kwargs):
        """
        Load training data from HDF5 file

        Parameters
        ----------
        **kwargs : if this contains file_path=<path_to_file>, load this file
        directly, otherwise open up a filedialog window

        Returns
        -------
        None.

        """

        if 'name' in kwargs:
            name = kwargs['name']
        else:
            name = 'Load training data'

        if 'file_path' in kwargs:
            file_path = kwargs['file_path']
        else:
            root = tk.Tk()
            root.withdraw()
            file_path = tk.filedialog.askopenfilename(title=name,
                                                      filetypes=(('HDF5 files', '*.hdf5'),
                                                                 ('All files', '*.*')))
        # load HDF5 data
        h5f = h5py.File(file_path, 'r')
        data_frame = {}

        # convert HDF5 data to a dictionary
        for key in h5f.keys():
            print('Loaded %s' % key)
            data_frame[key] = h5f[key][()]

        return data_frame

    def store_data_to_hdf5(self, data, **kwargs):
        """
        Store data to HDF5 format

        Parameters
        ----------
        data : dict
            data stored in a dictionary or h5py file object.

        file_path : sting (optional)
            A path to the file. If unspecified a file dialog is opened.

        Returns
        -------
        None.

        """

        if not isinstance(data, dict) and not isinstance(data, h5py.File):
            print('Data must be stored in a dict or h5py File object.')
            return

        if 'file_path' in kwargs:
            file_path = kwargs['file_path']
            file = open(file_path, 'wb')
        else:
            root = tk.Tk()
            root.withdraw()
            file = filedialog.asksaveasfile(title="Store data in HDF5 file",
                                            mode='wb', defaultextension=".hdf5")
        print('Saving data in', file.name)
        # create HDF5 file
        h5f = h5py.File(file, 'w')

        # store numpy sample arrays as individual datasets in the hdf5 file
        for name in data.keys():
            h5f.create_dataset(name, data=data[name])

        h5f.close()
        file.close()
        print('done')

    def accumulate_data(self, data, **kwargs):
        """
        Accumulate data over a number of (time) steps. Data is stored in
        a list in the self.accum_data dictionary, under the same key as
        given in the 'data' argument.

        Parameters
        ----------
        data : dict
            The key-worded data that is accumulated over a number of (time) steps.
        kwargs : keyword arguments
            If kwargs contains 'names', which is a list of keyword names, then
            the data is stored under the specified names rather than the names
            given by data.keys().

        Returns
        -------
        None.

        """

        if not isinstance(data, dict):
            print('Data must be stored in a dict.')
            return

        if 'names' in kwargs:
            names = kwargs['names']
        else:
            names = list(data.keys())
        idx = 0
        for key in data.keys():
            if not names[idx] in self.accum_data:
                self.accum_data[names[idx]] = []
            self.accum_data[names[idx]].append(data[key])
            idx += 1

    def store_accumulated_data(self, **kwargs):
        """
        Stores the data accumulated by the 'accumulate_data' subroutine to
        a HDF5 file.

        Returns
        -------
        None.

        """
        self.store_data_to_hdf5(self.accum_data, **kwargs)

    def add_app(self, name=None, surrogate=None):
        """
        Add an application.

        Parameters
        ----------
        name : string (optional)
            Name of the EasySurrogate campaign. If unspecified the self.campaign_name is used.
        surrogate : object
            A surrogate object.

        Returns
        -------
        None.

        """
        if name is None:
            name = self.campaign_name

        self.campaign_name = name
        self.surrogate = surrogate

    def load_state(self, **kwargs):
        """
        Loads the state of the current campaign from a pickle file

        Parameters
        ----------
        **kwargs : if this contains file_path=<path_to_file>, load this file
                   directly, otherwise open up a filedialog window
        Returns
        -------
        None.

        """

        if 'name' in kwargs:
            name = kwargs['name']
        else:
            name = ''

        if 'file_path' in kwargs:
            file_path = kwargs['file_path']
        else:
            root = tk.Tk()
            root.withdraw()

            file_path = filedialog.askopenfilename(title="Load state %s" % name,
                                                   filetypes=(('pickle files', '*.pickle'),
                                                              ('All files', '*.*')))

        print('Loading state from', file_path)

        file = open(file_path, 'rb')
        state = pickle.load(file)
        file.close()

        self.__dict__ = state

    def save_state(self, state=None, **kwargs):
        """
        Save the state of the current campaign to a pickle file

        Parameters
        ----------
        state: dict or None
            If None, store the state of the Campaign, not including the training data. This
            subroutine is inherited by the Surrogate classes to store their state.

        **kwargs :
            Can contain file_path=<path_to_file>, to save to this file  directly,
            otherwise open up a filedialog window.

        Returns
        -------
        None.

        """

        if state is None:
            state = self.__dict__

        if 'name' in kwargs:
            name = kwargs['name']
        else:
            name = self.campaign_name

        if 'file_path' in kwargs:
            file = open(kwargs['file_path'], 'wb')
        else:
            root = tk.Tk()
            root.withdraw()

            file = filedialog.asksaveasfile(title="Save state %s" % name,
                                            mode='wb',
                                            defaultextension=".pickle")

        print('Saving state to %s.' % file.name)
        pickle.dump(state, file)
        file.close()



"""
==============================================================================
CLASS FOR A ARTIFICIAL NEURAL NETWORK
------------------------------------------------------------------------------
Author: W. Edeling
==============================================================================
"""

import easysurrogate as es
from ..campaign import Campaign


class ANN_Surrogate(Campaign):
    """
    ANN_surrogate class for a standard feed forward neural network.
    """

    def __init__(self, **kwargs):
        print('Creating ANN_Surrogate Object')
        self.name = 'ANN Surrogate'
        # create a Feature_Engineering object
        self.feat_eng = es.methods.Feature_Engineering()

    ############################
    # START COMMON SUBROUTINES #
    ############################

    def train(self, feats, target, n_iter, lags=None, local=False,
              test_frac=0.0,
              n_layers=2, n_neurons=100,
              activation='tanh',
              batch_size=64, lamb=0.0, **kwargs):
        """
        Perform back propagation to train the ANN

        Parameters
        ----------
        feats : feature array, or list of different feature arrays
        target : the target data
        lags : list of time lags per feature
        n_iter : number of back propagation iterations
        test_frac : Fraction of the data to use for training.
                    The default is 0.0.
        n_layers : The number of layers in the neural network. Includes the
                   input layer and the hidden layers. The default is 2.
        n_neurons : The number of neurons per layer. The default is 100.
        activation : Type of activation function. The default is 'tanh'.
        batch_size : Mini batch size. The default is 64.
        lamb : L2 regularization parameter. The default is 0.0.

        Returns
        -------
        None.

        """

        # time lags
        self.lags = lags
        # flag if the surrogate is to be applied locally or not
        self.local = local

        # prepare the training data
        X_train, y_train = self.feat_eng.get_training_data(feats, target, lags=lags, local=local,
                                                           test_frac=test_frac)
        # get the maximum lag that was specified
        self.max_lag = self.feat_eng.max_lag

        # number of output neurons
        n_out = y_train.shape[1]

        # create the feed-forward ANN
        self.neural_net = es.methods.ANN(X=X_train, y=y_train,
                                         n_layers=n_layers, n_neurons=n_neurons,
                                         n_out=n_out,
                                         loss='squared',
                                         activation=activation, batch_size=batch_size,
                                         lamb=lamb, decay_step=10**4, decay_rate=0.9,
                                         standardize_X=True, standardize_y=True,
                                         save=False, **kwargs)

        print('===============================')
        print('Training Artificial Neural Network...')

        # train network for N_iter mini batches
        self.neural_net.train(n_iter, store_loss=True)
        self.set_data_stats()
        if lags is not None:
            self.feat_eng.initial_condition_feature_history(feats)

    def train_online(self, n_iter=1, batch_size=1, verbose=False, sequential=False):
        """
        Perform online training, i.e. backpropagation while the surrogate is coupled
        to the macroscopic governing equations.

        Source:
        Rasp, "Coupled online learning as a way to tackle instabilities and biases
        in neural network parameterizations: general algorithms and Lorenz 96
        case study", 2020.

        Parameters
        ----------
        feats : feature array, or list of different feature arrays
        target : the target data
        n_iter : number of back propagation iterations
        batch_size : Mini batch size. Default is 1.
        verbose: print loss to screen during back propagation. Default is False.
        sequential: do not randomly sample the training data, Default is False.

        Returns
        -------
        None.

        """

        X_train, y_train = self.feat_eng.get_online_training_data(n_in=self.neural_net.n_in,
                                                                  n_out=self.neural_net.n_out)

        # set the training data, training size and batch size for the online backprop step
        if self.neural_net.batch_size != batch_size:
            self.neural_net.set_batch_size(batch_size)
        self.neural_net.n_train = X_train.shape[0]
        # standardize training data
        self.neural_net.X = (X_train - self.feat_mean) / self.feat_std
        self.neural_net.y = (y_train - self.output_mean) / self.output_std

        # train network for n_iter mini batches
        self.neural_net.train(n_iter, store_loss=True, sequential=sequential, verbose=verbose)

    def generate_online_training_data(self, feats, LR_before, LR_after, HR_before, HR_after):
        """
        Compute the features and the target data for an online training step. Results are
        stored internally, and used within the 'train_online' subroutine.

        Source:
        Rasp, "Coupled online learning as a way to tackle instabilities and biases
        in neural network parameterizations: general algorithms and Lorenz 96
        case study", 2020.

        Parameters
        ----------
        feats : array or list of arrays
            The input features
        LR_before : array
            Low resolution state at previous time step.
        LR_after : array
            Low resolution state at current time step.
        HR_before : array
            High resolution state at previous time step.
        HR_after : array
            High resolution state at current time step.

        Returns
        -------
        None.

        """

        self.feat_eng.generate_online_training_data(feats, LR_before, LR_after, HR_before, HR_after)

    def set_online_training_parameters(self, tau_nudge, dt_LR, window_length):
        """
        Stores parameters required for online training.

        Parameters
        ----------
        tau_nudge : float
            Nudging time scale.
        dt_LR : float
            Time step low resolution model.
        window_length : int
            The length of the moving window in which online features are stored.

        Returns
        -------
        None.

        """
        self.feat_eng.set_online_training_parameters(tau_nudge, dt_LR, window_length)

    def predict(self, feat):
        """
        Make a prediction f(feat). Here, f is given by ANN_Surrogate_feed_foward.

        Parameters
        ----------
        feat : array of list of arrays
               The feature array of a list of feature arrays on which to evaluate the surrogate.

        Returns
        -------
        array
            the prediction of the neural net.

        """

        # feat_eng._predict handles the preparation of the features and returns
        # self._feed_forward(X)
        return self.feat_eng._predict(feat, self._feed_forward)

    def _feed_forward(self, feat):
        """
        A feed forward run of the ANN. This is the only part of prediction that is specific
        to ANN_Surrogate.

        Parameters
        ----------
        feat : array of list of arrays
               The feature array of a list of feature arrays on which to evaluate the surrogate.

        Returns
        -------
        y : array
            the prediction of the neural net.

        """

        # features were standardized during training, do so here as well
        feat = (feat - self.feat_mean) / self.feat_std
        # feed forward prediction step
        y = self.neural_net.feed_forward(feat.reshape([1, self.neural_net.n_in])).flatten()
        # transform y back to physical domain
        y = y * self.output_std + self.output_mean

        return y

    def save_state(self):
        """
        Save the state of the ANN surrogate to a pickle file
        """
        state = self.__dict__
        super().save_state(state=state, name=self.name)

    def load_state(self):
        """
        Load the state of the ANN surrogate from file
        """
        super().load_state(name=self.name)

    ##########################
    # END COMMON SUBROUTINES #
    ##########################

    def set_data_stats(self):
        """
        If the data were standardized, this stores the mean and
        standard deviation of the features in the ANN_Surrogate object.
        """
        if hasattr(self.neural_net, 'X_mean'):
            self.feat_mean = self.neural_net.X_mean
            self.feat_std = self.neural_net.X_std
        else:
            self.feat_mean = 0.0
            self.feat_std = 1.0

        if hasattr(self.neural_net, 'y_mean'):
            self.output_mean = self.neural_net.y_mean
            self.output_std = self.neural_net.y_std
        else:
            self.output_mean = 0.0
            self.output_std = 1.0



"""
Deep Active Subspace layer
"""
import numpy as np

from .Layer import Layer


class DAS_Layer(Layer):
    """
    Deep Active Subspace layer with weights parameterized by Gram-Schmidt
    orthogonalization. Place between input layer and first hidden layer.
    Inherits from the standard Layer class.

    Method:
        Tripathy, Rohit, and Ilias Bilionis. "Deep active subspaces: A scalable
        method for high-dimensional uncertainty propagation."
        ASME 2019 International Design Engineering Technical Conferences and
        Computers and Information in Engineering Conference.
        American Society of Mechanical Engineers Digital Collection, 2019.
    """

    def __init__(self, d, n_layers, bias, batch_size=1):
        """
        Initialize the DAS_layer oject.

        Parameters
        ----------
        d : int
            The dimension of the active subspace.
        bias : boolean, optional
            Add a bias neuron to the DAS layer. NOT IMPLEMENTED YET.
            The default is False.
        batch_size : int, optional
            The batch size. The default is 1.

        Returns
        -------
        None.

        """
        super().__init__(d, 1, n_layers, 'linear', 'none',
                         bias=bias, batch_size=batch_size)
        self.d = d
        self.grad_Phi = np.ones([self.n_neurons, self.batch_size])
        self.name = 'DAS_layer'

    def meet_the_neighbors(self, layer_rm1, layer_rp1):
        """
        Connect this layer to its neighbors

        Parameters
        ----------
        layer_rm1 : Layer object or None
            The layer before at index r - 1.
        layer_rp1 : Layer object or None
            The layer after at index r + 1.

        Returns
        -------
        None.

        """
        super().meet_the_neighbors(layer_rm1, layer_rp1)
        self.D = layer_rm1.n_neurons

    def init_weights(self):
        """
        Initialize the weights and othe related matrices of this DAS layer.

        Returns
        -------
        None.

        """
        # initialize the weights of the layer, which parameterize the
        # Gram-Schmidt vectors w
        self.Q = np.random.randn(self.layer_rm1.n_neurons, self.n_neurons) * \
            np.sqrt(1.0 / self.layer_rm1.n_neurons)
        # the unnormalized Gram-Schmidt vectors
        self.w = np.zeros([self.layer_rm1.n_neurons, self.n_neurons])
        # compute the Gram-Schmidt vectors, given Q
        self.compute_weights()
        # the gradient of the loss function
        self.L_grad_W = np.zeros([self.layer_rm1.n_neurons, self.n_neurons])
        # momentum
        self.V = np.zeros([self.layer_rm1.n_neurons, self.n_neurons])
        # squared gradient
        self.A = np.zeros([self.layer_rm1.n_neurons, self.n_neurons])

    def compute_weights(self):
        """
        Compute the Gram-Schmidt vectors w(Q)

        Returns
        -------
        None.

        """
        # This is a stadard Gram-Schmidt implementation, computing orthonormal
        # vectors from another set of vectors Q = [q1 q2 ... q_d].
        self.w[:, 0] = self.Q[:, 0]

        for i in range(1, self.d):
            self.w[:, i] = self.Q[:, i]
            for j in range(i):
                self.w[:, i] -= (np.dot(self.w[:, j], self.Q[:, i]) /
                                 np.dot(self.w[:, j], self.w[:, j])) * self.w[:, j]

        # the vectors w are not normalized, the weights of the neural network (W)
        # are normalized as w / norm(w)
        self.norm_w = np.linalg.norm(self.w, axis=0)
        self.W = self.w / self.norm_w

    def compute_L_grad_Q(self):
        """
        Compute the gradient of the loss function wrt the weights of this layer.
        Unlike standard layer, the weights are parameterized by Q, and so we
        must compute the gradient of the loss function with respect to Q here.

        Returns
        -------
        None.

        """
        # This part is standard back prop: compute L_grad_W
        h_rm1 = self.layer_rm1.h
        delta_ho_grad_Phi = self.delta_ho * self.grad_Phi
        self.L_grad_W = np.dot(h_rm1, delta_ho_grad_Phi.T)
        # here we compute W_grad_q_ij, the gardient of Q wrt each entry in Q
        # The results are stored in a 3D array of shape (Dd, D, d), which contains
        # the Dd matrices dW / dq_ij, i=1,...,D, j=1,...,d.
        self.W_grad_q_ij = self.compute_W_grad_q_ij()
        # This computes the gradient of loss function wrt the matrix Q via the
        # chain rule: dL / dQ = dL / dW * dW / dq_11 + ... + dl / dW * dW / dq_Dd
        self.L_grad_Q = np.sum(self.W_grad_q_ij * self.L_grad_W,
                               axis=(1, 2)).reshape([self.D, self.d])

    def compute_y_grad_Q(self):
        """
        Compute the gradient of the output wrt the weights of this layer.
        Unlike standard layer, the weights are parameterized by Q, and so we
        must compute the gradient of the loss function with respect to Q here.

        Returns
        -------
        None.

        """
        # here we compute W_grad_q_ij, the gardient of Q wrt each entry in Q
        # The results are stored in a 3D array of shape (Dd, D, d), which contains
        # the Dd matrices dW / dq_ij, i=1,...,D, j=1,...,d.
        self.W_grad_q_ij = self.compute_W_grad_q_ij()
        # This computes the gradient of loss function wrt the matrix Q via the
        # chain rule: dL / dQ = dL / dW * dW / dq_11 + ... + dl / dW * dW / dq_Dd
        self.y_grad_Q = np.sum(self.W_grad_q_ij * self.y_grad_W,
                               axis=(1, 2)).reshape([self.D, self.d])

    def compute_W_grad_q_ij(self):
        """
        Compute the gradient of the weights W (the normalized Gram-Schmidt vectors)
        with respect to every entry of the matrix Q.

        Returns
        -------
        dWdq_ij : array
            The Dd gradients dW / dq_ij, i=1,...,D, j=1,...,d. Stored in an array
            of shape (Dd, D, d).

        """

        d = self.d
        D = self.D
        # create a list of the column vectors of W and Q, with shape (D,1)
        w = [w_i.reshape([self.D, 1]) for w_i in list(self.w.T)]
        q = [q_i.reshape([self.D, 1]) for q_i in list(self.Q.T)]
        # gradients of the Gram-Schmidt vectors dw_i / dq_k
        grads = np.zeros([d**2, D, D])
        # gradients of the normed Gram-Schmidt vectos d(w_i / ||w_i||_2) / dq_k
        grads_normed = np.zeros([d**2, D, D])
        # D_ij matrices
        D_ij = np.zeros([d**2, D, D])
        # for w_1, the gradient and D_ij matrices have a simple form. Compute these
        # outside loop.
        I_D = np.eye(D)
        D_ij[0] = I_D
        grads[0] = I_D
        grads_normed[0] = I_D / self.norm_w[0] - np.dot(w[0], w[0].T) / self.norm_w[0]**3

        # the D_ij matrices must be precomputed
        # loop over dw_i, i = 2,...,d
        for i in range(1, d):
            # loop over all dq_j, j = 1,...,d
            for j in range(d):
                # convert 2D index (i,j) to scalar index idx
                idx = np.ravel_multi_index([i, j], dims=(d, d))
                if i >= j:
                    D_ij[idx] = compute_D_ij(w[j], q[i])

        # construct the d^2 gradient matrices dw_i / dq_k
        for i in range(1, d):
            # the matrix by which the gradient d_wi / dq_k must be premultiplied
            norm_mat = I_D / self.norm_w[i] - np.dot(w[i], w[i].T) / self.norm_w[i]**3
            for k in range(d):
                # convert 2D index (i,j) to scalar index idx
                idx = np.ravel_multi_index([i, k], dims=(d, d))
                # due to GS, the gradient dw_i / dq_k is zero for k > i, so compute if i >= k
                if i >= k:
                    # a 'normal' gradient dw_i / dq_i
                    if i == k:
                        # index of dw_{i-1} / dq_{i-1}
                        idx1 = np.ravel_multi_index([i - 1, i - 1], dims=(d, d))
                        grad_ik = grads[idx1] - \
                            np.dot(w[i - 1], w[i - 1].T) / np.dot(w[i - 1].T, w[i - 1])
                    # a 'shear' gradient dw_i / dq_k where i is not k
                    else:
                        grad_ik = 0.0
                        # loop over j=1,...,i - 1
                        for j in range(i):
                            # index of D_{ij}
                            idx2 = np.ravel_multi_index([i, j], dims=(d, d))
                            # index of dw_j / dq_k
                            idx3 = np.ravel_multi_index([j, k], dims=(d, d))
                            grad_ik -= np.dot(D_ij[idx2], grads[idx3])
                    # store dw_i / dq_k
                    grads[idx] = grad_ik
                    # compute the derivates of w_i / ||w_i||_2
                    grads_normed[idx] = np.dot(norm_mat, grad_ik)

        # re-order the gradient matrices to get get dW / dq_ij, i=1,...,D, j = 1,...,d.
        # Here, W = [w_1, w_2, ..., w_d] is the matrix of all orthogonal vectors
        dWdq_ij = np.zeros([D * d, D, d])
        counter = 0
        for i in range(D):
            for j in range(d):
                # selects d dw_k / dq_j matrices (k = 1,...,d and fixed j)
                idx = np.arange(j, d**2, d)
                # the gradient of W wrt scalar q_ij: select the i-th column of
                # the d dw_k / dq_j matrices (k = 1,...,d and fixed j)
                dWdq_ij[counter] = grads_normed[idx, :, i].T
                counter += 1

        return dWdq_ij

    def back_prop(self, y_i=None, jacobian=False):
        """
        Perform the backpropogation operations of the current layer. Unlike in the
        standard Layer, we compute the gradient with respect to the matrix Q, since
        the weights are parameterized by Q.

        Returns
        -------
        None.

        """
        # compute the standard gradient of the loss function wrt the activation
        # functions, uses method from parent class.
        self.compute_delta_ho()
        # compute the loss gradient wrt Q, since we must update Q
        self.compute_L_grad_Q()

        if jacobian:
            self.compute_delta_hy()
            self.compute_y_grad_W()
            self.compute_y_grad_Q()
            self.L_grad_Q = -self.y_grad_Q


def compute_D_ij(w_j, q_i):
    """
    Compute the D_ij matrices which appear in the gradients of the
    Gram-Schmidt vectors w_i wrt the original unnormalized vectors q_j.
    These matrices are only defined for i unequal to j.

    D_ij = 1/(w_j^Tw_j) [w_jq_i^T - (2w_j^Tq_i)/(w_j^Tw_j)*w_jw_j^T + w_j^Tq_i*I]
    where I is the D x D identity matrix.

    Parameters
    ----------
    w_j : array, shape (D,1)
        The orthogonal, but not yet normalized, vector w_i, i=1,...,d.
    q_i : array, shape (D,1)
        An orginal, non-orthonogal, vector q_j, j=1,...,d.

    Returns
    -------
    D_ij : array
        The D x D matrix described above.

    """
    return (np.dot(w_j, q_i.T) -
            (2.0 * np.dot(w_j.T, q_i) / np.dot(w_j.T, w_j)) * np.dot(w_j, w_j.T) +
            np.dot(w_j.T, q_i) * np.eye(q_i.size)) / np.dot(w_j.T, w_j)
